"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Circuit, ChannelType, Stabilizer, PauliOperator
from loom.eka.operations import Split
from loom.eka.utilities import Direction, Orientation
from loom.interpreter import InterpretationStep
from loom.interpreter.utilities import Cbit

from loom_rotated_surface_code.code_factory import RotatedSurfaceCode


# pylint: disable=duplicate-code
def split_consistency_check(
    interpretation_step: InterpretationStep, operation: Split
) -> RotatedSurfaceCode:
    """Check that the split operation can be performed on the given block.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block to split.
    operation : Split
        Split operation to perform.

    Returns
    -------
    RotatedSurfaceCode
        The block to split.

    Raises
    ------
    ValueError
        If the split position is larger than the width or height of the block.
    """
    block = interpretation_step.get_block(operation.input_block_name)

    if not isinstance(block, RotatedSurfaceCode):
        raise TypeError(
            f"The split operation is not supported for {type(block)} blocks."
        )
    position = operation.split_position
    if len(block.logical_x_operators) >= 2 or len(block.logical_z_operators) >= 2:
        raise NotImplementedError(
            "Splitting blocks with more than one logical operator is not supported."
        )
    if operation.orientation == Orientation.HORIZONTAL:
        if position >= block.size[1]:
            raise ValueError(
                f"Split position {position} is larger than the width of the block "
                f"({block.size[1]})."
            )
        if position in (0, block.size[1] - 1):
            raise ValueError("Split position cannot be at the edge of the block.")
    else:
        if position >= block.size[0]:
            raise ValueError(
                f"Split position {position} is larger than the height of the block "
                f"({block.size[0]})."
            )
        if position in (0, block.size[0] - 1):
            raise ValueError("Split position cannot be at the edge of the block.")

    return block


def create_split_circuit(
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    operation: Split,
    qubits_to_measure: tuple[tuple[int, ...], ...],
    boundary_type: str,
) -> tuple[Circuit, list[Cbit]]:
    """
    Create the circuit for the split operation.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block to split.
    block : RotatedSurfaceCode
        Block to split.
    operation : Split
        Split operation to perform.
    qubits_to_measure : tuple[tuple[int, ...], ...]
        Qubits to be measured in the split operation.
    boundary_type : str
        Type of the boundary of the block.

    Returns
    -------
    Circuit
        Circuit for the split operation.
    """
    # B) CIRCUIT
    #    B.1) Create classical channels for all data qubit measurements
    cbits = [interpretation_step.get_new_cbit_MUT(f"c_{q}") for q in qubits_to_measure]
    cbit_channels = [
        interpretation_step.get_channel_MUT(
            f"{cbit[0]}_{cbit[1]}", channel_type=ChannelType.CLASSICAL
        )
        for cbit in cbits
    ]
    #    B.2) Create a measurement circuit for every measured data qubit
    measure_circuit_seq = [
        [
            Circuit(
                "Measurement",
                channels=[interpretation_step.get_channel_MUT(q), cbit_channels[i]],
            )
            for i, q in enumerate(qubits_to_measure)
        ]
    ]

    #  If needed, apply a basis change
    if boundary_type == "X":
        # If the boundary type is X, the data qubit can directly be read out in the
        # Z basis
        split_circuit_list = measure_circuit_seq
    else:
        # If the boundary type is Z, the 2-bodies stabilizers are X, the data qubits
        # have to be read out in the X basis. Apply Hadamard gates before the
        # Z measurement to effectively measure in the X basis
        basis_change_circuit_seq = [
            [
                Circuit("H", channels=[interpretation_step.get_channel_MUT(q)])
                for q in qubits_to_measure
            ]
        ]
        # We add two sequences of gates to the circuit, i.e. 2 timesteps
        split_circuit_list = basis_change_circuit_seq + measure_circuit_seq

    row_column_str = (
        "row" if operation.orientation == Orientation.HORIZONTAL else "column"
    )
    split_circuit = Circuit(
        name=f"Split {block.unique_label} at {row_column_str} "
        f"{operation.split_position}",
        circuit=split_circuit_list,
    )
    return split_circuit, cbits


def split_stabilizers(
    block: RotatedSurfaceCode,
    operation: Split,
    qubits_to_measure: tuple[tuple[int, ...], ...],
) -> tuple[list[Stabilizer], list[Stabilizer], list[Stabilizer], list[Stabilizer]]:
    """
    Split the initial block's stabilizers into two lists of stabilizers, once for
    each block, and the list of old stabilizers that are gonna be replaced and the new
    stabilizers that will replace them.

    Parameters
    ----------
    block : RotatedSurfaceCode
        Initial block to be split
    operation : Split
        Split operation description
    qubits_to_measure : tuple[tuple[int, ...], ...]
        Qubits to be measured in the split operation.

    Returns
    -------
    tuple[list[Stabilizer], list[Stabilizer], list[Stabilizer], list[Stabilizer]]
        A tuple of four lists of stabilizers, one for each block, the old
        stabilizers that are gonna be replaced and the new stabilizers that will replace
        them (in the same order).
    """

    # C) - STABILIZERS
    #    C.1) Find stabilizers which are completely removed
    #    C.2) Find stabilizers which have to be reduced in weight
    #    C.3) Create new stabilizers with reduced weight
    #    C.4) Create two sets of stabilizers, one for each new block
    split_boundary = (
        Direction.LEFT
        if operation.orientation == Orientation.VERTICAL
        else Direction.TOP
    )
    boundary_type = block.boundary_type(split_boundary)
    split_is_vertical = operation.orientation == Orientation.VERTICAL
    split_position = operation.split_position

    #    C.1) Find stabilizers which are completely removed
    stabs_to_remove = [
        stab
        for stab in block.stabilizers
        if any(q in qubits_to_measure for q in stab.data_qubits)
    ]
    #    C.2) Find stabilizers which have to be reduced in weight
    old_stabs_to_reduce_weight = [
        stab
        for stab in stabs_to_remove
        if stab.pauli[0] != boundary_type
        and len([q for q in stab.data_qubits if q in qubits_to_measure]) == 2
        and len(stab.data_qubits) == 4
    ]
    #    C.3) Create new stabilizers with reduced weight
    new_stabs_reduced_weight = [
        Stabilizer(
            pauli="".join(
                stab.pauli[i]
                for i, q in enumerate(stab.data_qubits)
                if q not in qubits_to_measure
            ),
            data_qubits=[q for q in stab.data_qubits if q not in qubits_to_measure],
            ancilla_qubits=stab.ancilla_qubits,
        )
        for stab in old_stabs_to_reduce_weight
    ]

    new_stabilizers = [
        stab for stab in block.stabilizers if stab not in stabs_to_remove
    ] + new_stabs_reduced_weight

    data_qubits_block_1 = set(
        q
        for q in block.data_qubits
        if q[not split_is_vertical]
        < split_position + block.upper_left_qubit[not split_is_vertical]
    )
    data_qubits_block_2 = set(
        q
        for q in block.data_qubits
        if q[not split_is_vertical]
        > split_position + block.upper_left_qubit[not split_is_vertical]
    )
    #    C.4) Create two sets of stabilizers, one for each new block
    new_stabs_block_1 = [
        stab
        for stab in new_stabilizers
        if set(stab.data_qubits).issubset(data_qubits_block_1)
    ]
    new_stabs_block_2 = [
        stab
        for stab in new_stabilizers
        if set(stab.data_qubits).issubset(data_qubits_block_2)
    ]

    remaining_stabs = (
        set(new_stabilizers) - set(new_stabs_block_1) - set(new_stabs_block_2)
    )
    # Consistency check that all new stabilizers are assigned to a block
    if len(remaining_stabs) != 0:
        raise ValueError(
            f"Stabilizers {remaining_stabs} are not assigned to any block."
        )

    return (
        new_stabs_block_1,
        new_stabs_block_2,
        old_stabs_to_reduce_weight,
        new_stabs_reduced_weight,
    )


def find_split_stabilizer_to_circuit_mappings(
    block: RotatedSurfaceCode,
    new_block_stabilizers: list[Stabilizer],
    new_boundary_direction: Direction,
) -> dict[str, tuple[str, ...]]:
    """
    Finds the mapping between the stabilizers in the new block and the associated
    syndrome circuits.

    Parameters
    ----------
    block : RotatedSurfaceCode
        Initial block to be split
    new_block_stabilizers : list[Stabilizer]
        List of stabilizers in the one of the new blocks
    new_boundary_direction : Direction
        Direction of the the new boundary in the new block

    Returns
    -------
    dict[str, tuple[str, ...]]
        New mapping of stabilizers to syndrome circuits for the block.
    """
    stabilizer_to_circuit = block.stabilizer_to_circuit
    new_stabs_id = [stab.uuid for stab in new_block_stabilizers]
    # Copy the old mapping for stabilizers that are conserved
    new_stab_to_circ = {
        stab_id: circ_id
        for (stab_id, circ_id) in stabilizer_to_circuit.items()
        if stab_id in new_stabs_id
    }
    # Find the syndrome circuits that should be associated with the new boundary
    # stabilizers
    new_boundary_stabs = [
        stab
        for stab in new_block_stabilizers
        if stab.uuid not in new_stab_to_circ.keys()
    ]
    pauli_str = new_boundary_stabs[0].pauli
    new_boundary_circuit = next(
        synd_circ
        for synd_circ in block.syndrome_circuits
        if synd_circ.pauli == pauli_str
        if synd_circ.name
        == f"{new_boundary_direction.value.lower()}-{pauli_str.lower()}"
    )
    new_stab_to_circ.update(
        {stab.uuid: new_boundary_circuit.uuid for stab in new_boundary_stabs}
    )

    return new_stab_to_circ


def split_logical_operators(
    # pylint: disable=too-many-locals
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    qubits_to_measure: tuple[tuple[int, ...], ...],
    operation: Split,
    upleft_qubit_block_1: tuple[int, ...],
    upleft_qubit_block_2: tuple[int, ...],
) -> tuple[
    InterpretationStep,
    tuple[PauliOperator, PauliOperator],
    tuple[PauliOperator, PauliOperator],
    dict[str, tuple[Cbit, ...]],
    dict[str, tuple[Cbit, ...]],
]:
    """
    Finds the logical operators for the new blocks after the split operation. If an
    operator is partially measured, it will be split into two operators. If it is fully
    measured, it is shifted to the new block, adding stabilizers to the operator
    evolution.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block to split.
    block : RotatedSurfaceCode
        Block to split.
    qubits_to_measure : tuple[tuple[int, ...], ...]
        Qubits involved in the split operation.
    operation : Split
        Split operation description.
    upleft_qubit_block_1 : tuple[int, ...]
        Upper left qubit of the first new block.
    upleft_qubit_block_2 : tuple[int, ...]
        Upper left qubit of the second new block.

    Returns
    -------
    tuple[
        InterpretationStep,
        tuple[PauliOperator, PauliOperator],
        tuple[PauliOperator, PauliOperator]
        dict[str, tuple[Cbit, ...]],
        dict[str, tuple[Cbit, ...]],
    ]
        The interpretation step updated with the logical operator evolutions, logical X
        and logical Z operators for the first block, logical X and logical Z
        operators for the second block, updates for the logical X and Z operators.

    Raises
    ------
    NotImplementedError
        If the initial block encodes more than one logical qubit.
    """
    # D) LOGICAL OPERATORS

    # Find the stabilizer required to update the logical operator
    split_is_vertical = operation.orientation == Orientation.VERTICAL
    split_position = operation.split_position

    #    D.1) Find logical operators which are not partially measured
    # The logical operator is "preserved" if it is not partially measured
    # If it was located at the same position as the split, we move it to the new blocks
    preserved_log_x_ops = [
        op
        for op in block.logical_x_operators
        if not any(q in op.data_qubits for q in qubits_to_measure)
        or all(q in op.data_qubits for q in qubits_to_measure)
    ]
    #    D.2) Find logical operators which have to be split
    # The logical operator is split if it is partially measured
    split_log_x_ops = [
        op
        for op in block.logical_x_operators
        if any(q in op.data_qubits for q in qubits_to_measure)
        and not all(q in op.data_qubits for q in qubits_to_measure)
    ]
    if len(preserved_log_x_ops) != 0 and len(split_log_x_ops) != 0:
        raise NotImplementedError(
            "Splitting blocks with more than one logical X operator is not supported."
        )
    #    D.3) Create new logical operators
    # If the operator is split, construct two operators from it
    if len(split_log_x_ops) != 0:
        qubits_in_log_x_1 = [
            q
            for q in split_log_x_ops[0].data_qubits
            if q[not split_is_vertical]
            < split_position + block.upper_left_qubit[not split_is_vertical]
        ]
        new_log_x_op_1 = PauliOperator(
            "".join(split_log_x_ops[0].pauli[i] for i in range(len(qubits_in_log_x_1))),
            qubits_in_log_x_1,
        )

        qubits_in_log_x_2 = [
            q
            for q in split_log_x_ops[0].data_qubits
            if q[not split_is_vertical]
            > split_position + block.upper_left_qubit[not split_is_vertical]
        ]
        new_log_x_op_2 = PauliOperator(
            "".join(split_log_x_ops[0].pauli[i] for i in range(len(qubits_in_log_x_2))),
            qubits_in_log_x_2,
        )
        stabs_required_x_1 = []
        stabs_required_x_2 = []
        new_x_op_updates = {}
    # The operator is not partially measured, it may need to be displaced
    else:
        new_log_x_op_1, stabs_required_x_1 = (
            block.get_shifted_equivalent_logical_operator(
                initial_operator=preserved_log_x_ops[0],
                new_upleft_qubit=upleft_qubit_block_1,
            )
        )
        new_log_x_op_2, stabs_required_x_2 = (
            block.get_shifted_equivalent_logical_operator(
                initial_operator=preserved_log_x_ops[0],
                new_upleft_qubit=upleft_qubit_block_2,
            )
        )
        new_x_op_updates = {
            new_log_x_op_1.uuid: interpretation_step.retrieve_cbits_from_stabilizers(
                stabs_required_x_1, block
            ),
            new_log_x_op_2.uuid: interpretation_step.retrieve_cbits_from_stabilizers(
                stabs_required_x_2, block
            ),
        }

    # The logical operator is "preserved" if it is not partially measured
    # If it was located at the same position as the split, we move it to the new blocks
    preserved_log_z_ops = [
        op
        for op in block.logical_z_operators
        if not any(q in op.data_qubits for q in qubits_to_measure)
        or all(q in op.data_qubits for q in qubits_to_measure)
    ]
    # The logical operator is split if it is partially measured
    split_log_z_ops = [
        op
        for op in block.logical_z_operators
        if any(q in op.data_qubits for q in qubits_to_measure)
        and not all(q in op.data_qubits for q in qubits_to_measure)
    ]
    if len(preserved_log_z_ops) != 0 and len(split_log_z_ops) != 0:
        raise NotImplementedError(
            "Splitting blocks with more than one logical Z operator is not supported."
        )
    if len(split_log_z_ops) != 0:
        qubits_in_log_z_1 = [
            q
            for q in split_log_z_ops[0].data_qubits
            if q[not split_is_vertical]
            < split_position + block.upper_left_qubit[not split_is_vertical]
        ]
        new_log_z_op_1 = PauliOperator(
            "".join(split_log_z_ops[0].pauli[i] for i in range(len(qubits_in_log_z_1))),
            qubits_in_log_z_1,
        )
        qubits_in_log_z_2 = [
            q
            for q in split_log_z_ops[0].data_qubits
            if q[not split_is_vertical]
            > split_position + block.upper_left_qubit[not split_is_vertical]
        ]
        new_log_z_op_2 = PauliOperator(
            "".join(split_log_z_ops[0].pauli[i] for i in range(len(qubits_in_log_z_2))),
            qubits_in_log_z_2,
        )
        stabs_required_z_1 = []
        stabs_required_z_2 = []
        new_z_op_updates = {}
    else:
        new_log_z_op_1, stabs_required_z_1 = (
            block.get_shifted_equivalent_logical_operator(
                initial_operator=preserved_log_z_ops[0],
                new_upleft_qubit=upleft_qubit_block_1,
            )
        )
        new_log_z_op_2, stabs_required_z_2 = (
            block.get_shifted_equivalent_logical_operator(
                initial_operator=preserved_log_z_ops[0],
                new_upleft_qubit=upleft_qubit_block_2,
            )
        )
        new_z_op_updates = {
            new_log_z_op_1.uuid: interpretation_step.retrieve_cbits_from_stabilizers(
                stabs_required_z_1, block
            ),
            new_log_z_op_2.uuid: interpretation_step.retrieve_cbits_from_stabilizers(
                stabs_required_z_2, block
            ),
        }

    #    D.4) Update `logical_x/z_evolution`
    # Hardcoded index for the logical operators !!!
    interpretation_step.logical_x_evolution[new_log_x_op_1.uuid] = tuple(
        [block.logical_x_operators[0].uuid] + [stab.uuid for stab in stabs_required_x_1]
    )
    interpretation_step.logical_x_evolution[new_log_x_op_2.uuid] = tuple(
        [block.logical_x_operators[0].uuid] + [stab.uuid for stab in stabs_required_x_2]
    )
    interpretation_step.logical_z_evolution[new_log_z_op_1.uuid] = tuple(
        [block.logical_z_operators[0].uuid] + [stab.uuid for stab in stabs_required_z_1]
    )
    interpretation_step.logical_z_evolution[new_log_z_op_2.uuid] = tuple(
        [block.logical_z_operators[0].uuid] + [stab.uuid for stab in stabs_required_z_2]
    )

    return (
        interpretation_step,
        (new_log_x_op_1, new_log_z_op_1),
        (new_log_x_op_2, new_log_z_op_2),
        new_x_op_updates,
        new_z_op_updates,
    )


def split(  # pylint: disable=line-too-long, too-many-locals
    interpretation_step: InterpretationStep,
    operation: Split,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Applicator to split a block using the Split description.

    The algorithm is the following:

    - A.) DATA QUBITS
    
        - Find data qubits to be measured, the initial block will be split into two

    - B.) CIRCUIT
    
        - B.1) Create classical channels for all data qubit measurements
        - B.2) Create a measurement circuit for every measured data qubit
        - B.3) Append the measurement circuits to the InterpretationStep circuit. \
        If needed, apply a basis change

    - C.) STABILIZERS
    
        - C.1) Find stabilizers which are completely removed
        - C.2) Find stabilizers which have to be reduced in weight
        - C.3) Create new stabilizers with reduced weight
        - C.4) Create two sets of stabilizers, one for each new block
        - C.5) Update ``stabilizer_evolution`` and ``stabilizer_updates`` for the \
        stabilizers which have been reduced in weight
        - C.6) Create the new ``stabilizer_to_circuit`` mapping

    - D.) LOGICAL OPERATORS
    
        - D.1) Find logical operators which are not partially measured
        - D.2) Find logical operators which have to be split
        - D.3) Create new logical operators and final to initial mapping
        - D.4) Update ``logical_x/z_evolution``
        - D.5) Update ``logical_x/z_updates``

    - E.) NEW BLOCK AND NEW INTERPRETATION STEP
    
        - E.1) Create the new blocks
        - E.2) Update the block history

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block to split.
    operation : Split
        Split operation description.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block

    Returns
    -------
    InterpretationStep
        Interpretation step after the split operation.
    """

    # Check that the operation can be performed
    block = split_consistency_check(interpretation_step, operation)
    split_is_vertical = operation.orientation == Orientation.VERTICAL
    split_position = operation.split_position

    # A) - DATA QUBITS
    #    Find data qubits to be measured, the initial block will be split into two
    split_boundary = (
        Direction.LEFT
        if operation.orientation == Orientation.VERTICAL
        else Direction.TOP
    )
    boundary_qubits = block.boundary_qubits(split_boundary)
    boundary_type = block.boundary_type(split_boundary)
    shift_vector = (
        split_is_vertical * split_position,
        (not split_is_vertical) * split_position,
        0,
    )
    qubits_to_measure = tuple(
        tuple(
            coord1 + coord2 for coord1, coord2 in zip(qubit, shift_vector, strict=True)
        )
        for qubit in boundary_qubits
    )

    # block 1 is either the left or top block
    qubits_block_1 = [
        q
        for q in block.data_qubits
        if q[not split_is_vertical]
        < split_position + block.upper_left_qubit[not split_is_vertical]
    ]
    upleft_qubit_block_1 = min(qubits_block_1, key=lambda x: x[0] + x[1])
    # block 2 is either the right or bottom block
    qubits_block_2 = [
        q
        for q in block.data_qubits
        if q[not split_is_vertical]
        > split_position + block.upper_left_qubit[not split_is_vertical]
    ]
    upleft_qubit_block_2 = min(qubits_block_2, key=lambda x: x[0] + x[1])

    # B) - CIRCUIT
    #    B.1) Create classical channels for all data qubit measurements
    #    B.2) Create a measurement circuit for every measured data qubit
    split_circuit, cbits = create_split_circuit(
        interpretation_step, block, operation, qubits_to_measure, boundary_type
    )
    #    B.3) Append the measurement circuits to the InterpretationStep circuit
    interpretation_step.append_circuit_MUT(split_circuit, same_timeslice)

    # C) - STABILIZERS
    #    C.1) Find stabilizers which are completely removed
    #    C.2) Find stabilizers which have to be reduced in weight
    #    C.3) Create new stabilizers with reduced weight
    #    C.4) Create two sets of stabilizers, one for each new block
    (
        new_stabs_block_1,
        new_stabs_block_2,
        old_stabs_to_reduce_weight,
        new_stabs_reduced_weight,
    ) = split_stabilizers(block, operation, qubits_to_measure)

    #    C.5) Update `stabilizer_evolution` and `stabilizer_updates` for the
    #         stabilizers which have been reduced in weight
    for i, stab in enumerate(new_stabs_reduced_weight):
        previous_updates = (
            interpretation_step.stabilizer_updates[stab.uuid]
            if stab.uuid in interpretation_step.stabilizer_updates.keys()
            else ()
        )
        old_stab = old_stabs_to_reduce_weight[i]
        # Find the data qubits of this stabilizer which are measured and whose
        # corresponding cbit has to be included in the stabilizer update
        qubits_measured = [q for q in old_stab.data_qubits if q in qubits_to_measure]
        # Find the cbits and include them in `stabilizer_updates`
        cbit_indices = [qubits_to_measure.index(q) for q in qubits_measured]
        new_stab_updates = tuple(cbits[cbit_idx] for cbit_idx in cbit_indices)
        if updates := previous_updates + new_stab_updates:
            interpretation_step.stabilizer_updates[stab.uuid] = updates

    stab_map_weight4_to_weight2 = {
        new_stabs_reduced_weight[i].uuid: (stab.uuid,)
        for i, stab in enumerate(old_stabs_to_reduce_weight)
    }
    interpretation_step.stabilizer_evolution.update(stab_map_weight4_to_weight2)

    #   C.6) Create the new `stabilizer_to_circuit` mapping
    # block 1 is the left block -> new boundary is right if split is vertical
    # block 1 is the top block -> new boundary is bottom if split is horizontal
    stab_to_circ_block_1 = find_split_stabilizer_to_circuit_mappings(
        block,
        new_stabs_block_1,
        Direction.RIGHT if split_is_vertical else Direction.BOTTOM,
    )
    # block 2 is the right block -> new boundary is left if split is vertical
    # block 2 is the bottom block -> new boundary is top if split is horizontal
    stab_to_circ_block_2 = find_split_stabilizer_to_circuit_mappings(
        block,
        new_stabs_block_2,
        Direction.LEFT if split_is_vertical else Direction.TOP,
    )

    # D) LOGICAL OPERATORS
    #    D.1) Find logical operators which are not partially measured
    #    D.2) Find logical operators which have to be split
    #    D.3) Create new logical operators and final to initial mapping
    #    D.4) Update `logical_x/z_evolution`
    (
        interpretation_step,
        (new_log_x_op_1, new_log_z_op_1),
        (new_log_x_op_2, new_log_z_op_2),
        new_x_updates,
        new_z_updates,
    ) = split_logical_operators(
        interpretation_step,
        block,
        qubits_to_measure,
        operation,
        upleft_qubit_block_1,
        upleft_qubit_block_2,
    )
    #    D.5) Create the new `logical_x/z_operator_updates`
    # Find measurements that split the relevant operator, if the operator is aligned
    # with the split, it's either fully measured or not measured at all.
    # It's not contributing to the operator updates in both cases.
    x_and_split_aligned = block.x_boundary == operation.orientation
    x_op_measurements = (
        tuple(
            c
            for c in cbits
            if c[0].split("_")[1] in map(str, block.logical_x_operators[0].data_qubits)
        )
        if not x_and_split_aligned
        else ()
    )
    z_op_measurements = (
        tuple(
            c
            for c in cbits
            if c[0].split("_")[1] in map(str, block.logical_z_operators[0].data_qubits)
        )
        if x_and_split_aligned
        else ()
    )
    # The new operator is mapped to the measurement that split the initial logical operator
    # The first block always inherits the updates from the initial block if the operator is modified
    interpretation_step.update_logical_operator_updates_MUT(
        operator_type="X",
        logical_operator_id=new_log_x_op_1.uuid,
        new_updates=new_x_updates.get(new_log_x_op_1.uuid, ()) + x_op_measurements,
        inherit_updates=(new_log_x_op_1.uuid != block.logical_x_operators[0].uuid),
    )
    interpretation_step.update_logical_operator_updates_MUT(
        operator_type="Z",
        logical_operator_id=new_log_z_op_1.uuid,
        new_updates=new_z_updates.get(new_log_z_op_1.uuid, ()) + z_op_measurements,
        inherit_updates=(new_log_z_op_1.uuid != block.logical_z_operators[0].uuid),
    )
    interpretation_step.update_logical_operator_updates_MUT(
        operator_type="X",
        logical_operator_id=new_log_x_op_2.uuid,
        new_updates=new_x_updates.get(new_log_x_op_2.uuid, ()),
        inherit_updates=x_and_split_aligned
        and (new_log_x_op_2.uuid != block.logical_x_operators[0].uuid),
    )
    interpretation_step.update_logical_operator_updates_MUT(
        operator_type="Z",
        logical_operator_id=new_log_z_op_2.uuid,
        new_updates=new_z_updates.get(new_log_z_op_2.uuid, ()),
        inherit_updates=not x_and_split_aligned
        and (new_log_z_op_2.uuid != block.logical_z_operators[0].uuid),
    )

    # E) NEW BLOCK AND NEW INTERPRETATION STEP
    #    E.1) Create the new blocks
    new_block_1 = RotatedSurfaceCode(
        stabilizers=new_stabs_block_1,
        logical_x_operators=[new_log_x_op_1],
        logical_z_operators=[new_log_z_op_1],
        syndrome_circuits=block.syndrome_circuits,
        stabilizer_to_circuit=stab_to_circ_block_1,
        unique_label=operation.output_blocks_name[0],
        skip_validation=not debug_mode,
    )
    new_block_2 = RotatedSurfaceCode(
        stabilizers=new_stabs_block_2,
        logical_x_operators=[new_log_x_op_2],
        logical_z_operators=[new_log_z_op_2],
        syndrome_circuits=block.syndrome_circuits,
        stabilizer_to_circuit=stab_to_circ_block_2,
        unique_label=operation.output_blocks_name[1],
        skip_validation=not debug_mode,
    )
    #    E.2) Update the block history
    interpretation_step.update_block_history_and_evolution_MUT(
        new_blocks=(new_block_1, new_block_2), old_blocks=(block,)
    )

    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from pydantic.dataclasses import dataclass

from ..eka import Block, LogicalState
from ..eka.utilities import (
    is_subset_of_stabarray,
    StabArray,
    SignedPauliOp,
    reduce_stabarray,
)
from ..cliffordsim import Engine
from ..cliffordsim.operations import UpdateTableau, Operation

from .check_abstract import AbstractValidityCheck
from .utilities import get_parity_from_cbits, all_possible_pauli_strings


@dataclass(frozen=True)
class LogicalOperatorCheckOutput:
    """
    Dataclass to store the output of the Logical Operator check.

    Parameters
    ----------
    input_vs_expected_vs_actual_logicals_multi: tuple[
        tuple[LogicalState, tuple[LogicalState, ...], LogicalState | None], ...
    ]
        A tuple containing the input logical states, the expected output logical states,
        and the actual output logical state. If the actual output state is None, it
        means that no valid logical state was found in the output.
    input_vs_expected_vs_actual_logicals_with_parity: tuple[
        tuple[LogicalState, tuple[LogicalState, tuple[int, ...]], LogicalState | None],
        ...
    ]
        A tuple containing the input logical states, the expected output logical states
        with parity flips, and the actual output logical state. If the actual output
        state is None, it means that no valid logical state was found in the output.

    """

    input_vs_expected_vs_actual_logicals_multi: tuple[
        tuple[LogicalState, tuple[LogicalState, ...], LogicalState | None], ...
    ]
    input_vs_expected_vs_actual_logicals_with_parity: tuple[
        tuple[LogicalState, tuple[LogicalState, tuple[int, ...]], LogicalState | None],
        ...,
    ]

    def __len__(self) -> int:
        """Returns the number of failed logical state transformations."""
        return len(self.input_vs_expected_vs_actual_logicals_multi) + len(
            self.input_vs_expected_vs_actual_logicals_with_parity
        )

    def __str__(self):
        out = ""
        if self.input_vs_expected_vs_actual_logicals_multi:
            out += "- Failed Logical State Transformations:\n"
            for (
                initial,
                allowed_final,
                actual,
            ) in self.input_vs_expected_vs_actual_logicals_multi:
                out += (
                    f"Initial: {initial}, Allowed Finals: {allowed_final},"
                    f" Actual: {actual}\n"
                )
        if self.input_vs_expected_vs_actual_logicals_with_parity:
            out += "- Expected vs Actual Logical Transformations:\n"
            for (
                initial,
                (expected, flips),
                actual,
            ) in self.input_vs_expected_vs_actual_logicals_with_parity:
                out += (
                    f"Initial: {initial}, Expected: {expected}, Flips: {flips},"
                    f" Actual: {actual}\n"
                )
        return out.rstrip("\n")  # Remove trailing newline for cleaner output


@dataclass(frozen=True)
class LogicalOperatorCheck(AbstractValidityCheck):
    """Dataclass to store the results of the Logical Operator check.

    Parameters
    ----------
    output: LogicalOperatorCheckOutput
        An object containing the output of the check, including any failed logical state
        transformations and the expected vs actual logical transformations with parity
        flips for each logical state.

    Properties
    ----------
    message: str
        A message indicating the result of the check. It will be empty if the check is
        valid, otherwise it will contain a message describing the issue.
    valid: bool
        True if the check is valid (i.e. all logical states were transformed
        correctly), False otherwise.
    """

    # Define allowed messages
    output: LogicalOperatorCheckOutput

    @property
    def message(self) -> str:
        if not self.valid:
            return "Some logical states were not transformed correctly."
        return ""


def check_logical_operators_transformation(
    base_cliffordsim_operations: tuple[Operation, ...],
    input_block: Block,
    output_block: Block,
    logical_state_transformations_with_parity: dict[
        LogicalState,
        tuple[LogicalState, dict[int, tuple[str | int, ...]]],
    ],
    logical_state_transformations: list[tuple[LogicalState, tuple[LogicalState, ...]]],
    seed: int | None,
) -> LogicalOperatorCheck:
    """Checks whether the logical operators are transformed correctly.

    Parameters
    ----------
    base_cliffordsim_operations : tuple[ \
        :class:`loom.cliffordsim.operations.base_operation.Operation`, ...]
        A tuple of base cliffordsim operations that will be used to run the circuit.
    input_block : Block
        The Block object that represents the input code.
    output_block : Block
        The Block object that represents the output code.
    logical_state_transformations_with_parity : dict[
        LogicalState,
        tuple[LogicalState, dict[int, tuple[str | int, ...]]],
    ]
        Dictionary where the keys are the input logical states and the values are
        tuples containing the output logical state and a dictionary of parity flips
        that correspond to each logical operator. The keys of the dictionary are the
        logical operator indices, and the values are lists of strings or integers (0 or
        1) that represent the classical channels and parity flips applied to the logical
        operators.
    logical_state_transformations : list[tuple[LogicalState, tuple[LogicalState, ...]]]
        A list of tuples where each tuple contains an input logical state and a tuple
        of expected output logical states. When the circuit is run for a given input
        logical state, at least one of the expected output logical states should be
        found in the output.
    seed : int | None
        The seed for the cliffordsim engine.

    Returns
    -------
    LogicalOperatorCheck
        The result of the Logical Operator check.
    """

    input_vs_expected_vs_actual_logicals_multi: list[
        tuple[LogicalState, tuple[LogicalState, ...], LogicalState | None]
    ] = []

    for input_state, output_states in logical_state_transformations:
        tableau = input_state.get_tableau(input_block)
        # Run the circuit from initial state
        operations = (UpdateTableau(tableau), *base_cliffordsim_operations)
        cliffordsim_engine = Engine(operations, input_block.n_data_qubits, seed=seed)
        cliffordsim_engine.run()

        # Retrieve the output stabilizer array
        out_stabarray = StabArray(cliffordsim_engine.tableau_w_scratch.stabilizer_array)

        # Get all possible valid output logical stabarrays
        expected_output_log_op_stabarray = [
            LogicalState.convert_logical_to_base_representation(
                output_block, output_state.stabarray
            )
            for output_state in output_states
        ]
        # Check that at least one of the expected logical operators is a subset of the
        # output StabArray
        if not any(
            is_subset_of_stabarray(exp_log_op_stabarr, out_stabarray)
            for exp_log_op_stabarr in expected_output_log_op_stabarray
        ):
            actual_output_state = find_logical_state(output_block, out_stabarray)

            input_vs_expected_vs_actual_logicals_multi.append(
                (
                    input_state,
                    output_states,
                    actual_output_state,
                )
            )

    input_vs_expected_vs_actual_logicals_with_parity: list[
        tuple[LogicalState, tuple[LogicalState, tuple[int, ...]], LogicalState | None]
    ] = []
    for input_state, (
        expected_output_state,
        parity_dict,
    ) in logical_state_transformations_with_parity.items():
        tableau = input_state.get_tableau(input_block)
        # Run the circuit from initial state
        operations = (UpdateTableau(tableau), *base_cliffordsim_operations)
        cliffordsim_engine = Engine(operations, input_block.n_data_qubits, seed=seed)
        cliffordsim_engine.run()

        # Retrieve the output stabilizer array
        out_stabarray = StabArray(cliffordsim_engine.tableau_w_scratch.stabilizer_array)

        # Get all possible valid output logical stabarrays
        exp_output_log_op_stabarray_without_parity = (
            LogicalState.convert_logical_to_base_representation(
                output_block, expected_output_state.stabarray
            )
        )

        # Create a list of lists of cbits for each logical operator
        # If the parity for a logical operator is not specified, it defaults to an
        # empty list
        cbits_list = [
            parity_dict.get(i, []) for i in range(output_block.n_logical_qubits)
        ]
        runtime_parity_list = [
            get_parity_from_cbits(cliffordsim_engine, cbits) for cbits in cbits_list
        ]

        # Find the expected output logical operator with the correct parities
        expected_output_log_op_stabarray = StabArray.from_signed_pauli_ops(
            [
                (op if runtime_parity == 0 else op.with_flipped_sign())
                for op, runtime_parity in zip(
                    exp_output_log_op_stabarray_without_parity,
                    runtime_parity_list,
                    strict=True,
                )
            ]
        )

        # Check that at least one of the expected logical operators is a subset of the
        # output StabArray
        if not is_subset_of_stabarray(expected_output_log_op_stabarray, out_stabarray):
            actual_output_state = find_logical_state(output_block, out_stabarray)

            input_vs_expected_vs_actual_logicals_with_parity.append(
                (
                    input_state,
                    (
                        expected_output_state,
                        tuple(runtime_parity_list),
                    ),
                    actual_output_state,
                )
            )

    return LogicalOperatorCheck(
        LogicalOperatorCheckOutput(
            input_vs_expected_vs_actual_logicals_multi=tuple(
                input_vs_expected_vs_actual_logicals_multi
            ),
            input_vs_expected_vs_actual_logicals_with_parity=tuple(
                input_vs_expected_vs_actual_logicals_with_parity
            ),
        )
    )


def find_logical_state(
    block: Block, output_stabarray: StabArray
) -> LogicalState | None:
    """
    Finds the logical state from the output stabilizer array. If the output stabilizer
    array does not contain a full set of logical operators, the logical state is not
    well-defined and None is returned.

    Parameters
    ----------
    block : Block
        The Block object that represents the code.
    output_stabarray : StabArray
        The stabilizer array from the output of the circuit.

    Returns
    -------
    LogicalState | None
        The logical state if it is well-defined, otherwise None.
    """

    log_stabarr_reduced = StabArray.trivial()
    for p_string in all_possible_pauli_strings(block.n_logical_qubits):
        # Convert the logical pauli string to a SignedPauliOp
        p_op = SignedPauliOp.from_string(p_string)

        # Convert the logical Pauli operator to its base representation
        # i.e. the logical Pauli operator is expressed in terms of block data qubits
        base_repr_p_op = LogicalState.convert_logical_pauli_op_to_base_representation(
            p_op, block.x_log_stabarray, block.z_log_stabarray
        )

        # Check if the base representation of the logical Pauli operator is a subset of
        # the output stabilizer array
        if is_subset_of_stabarray(base_repr_p_op, output_stabarray):
            # Append the operator to the reduced stabarray and reduce it again
            log_stabarr_reduced = reduce_stabarray(
                StabArray.from_signed_pauli_ops(list(log_stabarr_reduced) + [p_op])
            )

            # If the reduced logical stabarray has the same number of stabilizers as
            # the logical qubits of the block, we have found the logical state.
            # This means that the logical state is well-defined.
            if log_stabarr_reduced.nstabs == block.n_logical_qubits:
                return LogicalState.from_stabarray(log_stabarr_reduced)

    # If no well-defined logical state was found, return None.
    return None



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from uuid import uuid4
from typing import Literal
from pydantic.dataclasses import dataclass, Field

#: A Cbit denotes the location of a classical bit by a tuple of a string and an integer.
#: str: Name/id of classical register, int: measurement index inside this register
Cbit = tuple[str, int] | Literal[1, 0]


@dataclass
class CompositeOperationSession:
    """
    Class representing a composite operation session.

    Attributes:
    ------------
    start_timeslice_index : int
        The index of the timeslice where the composite operation session starts.
    same_timeslice : bool
        A flag indicating whether all operations in the composite operation are
        to be executed in the same timeslice as the previous timeslice.
    circuit_name : str
        The name assigned to the composite circuit.
    uuid : str
        A unique identifier for the composite operation session.
    """

    start_timeslice_index: int
    same_timeslice: bool
    circuit_name: str
    uuid: str = Field(
        default_factory=lambda: str(uuid4()), validate_default=True, init=False
    )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations
from itertools import combinations
from typing import Sequence
from functools import reduce
from copy import deepcopy
import re
import numpy as np

from .exceptions import AntiCommutationError
from .pauli_array import PauliArray
from .pauli_array_computation import rowsum
from .pauli_binary_vector_rep import SignedPauliOp, pauliops_anti_commute
from .graph_matrix_utils import binary_gaussian_elimination
from .tableau import tableau_generates_pauli_group


# pylint: disable=too-many-lines
class StabArray(PauliArray):
    """
    A StabArray is a collection of commuting SignedPauliOps acting on the same qubit
    register.

    Parameters
    ----------
    array : np.ndarray
        The numpy array containing the stabilizers.
    validated : bool, optional
        Whether the input array has already been validated, by default False.

    Attributes
    ----------
    nqubits : int
        The number of qubits that the PauliArray operators act on.
    x : np.ndarray
        The array representing the X-component of the PauliArray in binary
        representation.
    z : np.ndarray
        The array representing the Z-component of the PauliArray in binary
        representation.
    """

    array: np.ndarray
    nqubits: int
    DTYPE = SignedPauliOp.DTYPE

    def __init__(self, array: np.ndarray, validated: bool = False) -> None:
        """
        Initialization of a StabArray.

        Parameters
        ----------
        array : np.ndarray
            The numpy array containing the stabilizers.
        validated : bool, optional
            Whether the input array has already been validated, by default False.

        Raises
        ------
        TypeError
            When the input is not a numpy array.
        ValueError
            When the input array is not 2-dimensional or contains values other than 0
            and 1.
        AntiCommutationError
            When the input array contains anti-commuting operators.
        """
        if not validated:
            if not isinstance(array, np.ndarray):
                raise TypeError("The input has to be a numpy array.")
                # find the SignedPauliOp's that make up the StabArray

            if not array.ndim == 2:
                raise ValueError("The numpy array has to be 2 dimensional")

            if np.any(array * (array - 1)):
                raise ValueError("The array should only contain 0s and 1s.")

            if array.dtype != SignedPauliOp.DTYPE:
                # convert the array to the correct dtype
                array = array.astype(SignedPauliOp.DTYPE)

            nstabs = array.shape[0]
            # check for anti-commuting pairs of operators
            if any(
                pauliops_anti_commute(
                    SignedPauliOp(array[i], validated=True),
                    SignedPauliOp(array[j], validated=True),
                )
                for i in range(nstabs)
                for j in range(i + 1, nstabs)
            ):
                raise AntiCommutationError(
                    "StabArray should only contain commuting operators."
                )

        self.array = array

    def copy(self) -> StabArray:
        """
        Returns a deep copy of the StabArray object.

        Returns
        -------
        SignedPauliOp
            Deep copy of the StabArray object
        """
        return deepcopy(self)

    @classmethod
    def from_signed_pauli_ops(
        cls, signed_pauli_ops: Sequence[SignedPauliOp], validated: bool = False
    ) -> StabArray:
        """
        Create a StabArray from a sequence of SignedPauliOps.

        Parameters
        ----------
        signed_pauli_ops : Sequence[SignedPauliOp]
            The sequence of SignedPauliOps to be converted to a StabArray.

        Returns
        -------
        StabArray
            The StabArray created from the sequence of SignedPauliOps.

        Raises
        ------
        TypeError
            When the input sequence contains elements that are not of type
            SignedPauliOp.
        ValueError
            When the input sequence contains elements with different numbers of qubits.
        """
        if any(not isinstance(p, SignedPauliOp) for p in signed_pauli_ops):
            raise TypeError(
                "All elements of the input sequence should be of type SignedPauliOp."
            )

        if any(
            p1.nqubits != p2.nqubits
            for p1, p2 in zip(signed_pauli_ops, signed_pauli_ops[1:], strict=False)
        ):
            raise ValueError(
                "All elements of the input sequence should have the same number of"
                " qubits."
            )

        nstabs = len(signed_pauli_ops)
        ncols = 2 * signed_pauli_ops[0].nqubits + 1 if nstabs > 0 else 0

        array = np.zeros((nstabs, ncols), dtype=SignedPauliOp.DTYPE)
        for i, op in enumerate(signed_pauli_ops):
            array[i, :] = op.array

        return cls(array, validated=validated)

    @classmethod
    def trivial(cls) -> StabArray:
        """
        Return a trivial StabArray.
        """
        return StabArray(np.zeros((0, 0), dtype=cls.DTYPE), validated=True)

    @property
    def nstabs(self) -> int:
        """
        The number of stabilizers in the StabArray.
        """
        return self.array.shape[0]

    @property
    def is_trivial(self) -> bool:
        """
        Checks whether the StabArray is trivial, i.e. if it contains no stabilizers.
        """
        return self.nstabs == 0

    @property
    def is_irreducible(self) -> bool:
        """
        Checks whether the StabArray is irreducible, i.e. if it contains no stabilizers
        that can be generated as a product of other stabilizers.
        """
        return reduce_stabarray(self).nstabs == self.nstabs

    @property
    def as_paulistrings(self) -> list[str]:
        """
        Representation of Pauli operators in the StabArray as a list of strings.
        """
        return [str(pauliop) for pauliop in self]

    def __getitem__(self, key: int | slice):
        """
        Overloads the [] operator to return a SignedPauliOp or a list of
        SignedPauliOp.
        """
        # Return a list of SignedPauliOp
        if isinstance(key, slice):
            indices = range(*key.indices(self.nstabs))
            return [SignedPauliOp(self.array[i, :], validated=True) for i in indices]
        if isinstance(key, np.ndarray) and key.ndim == 1 and key.dtype == np.bool_:
            # Support for boolean indexing
            return [
                SignedPauliOp(self.array[i, :], validated=True)
                for i in range(self.nstabs)
                if key[i]
            ]
        # or a single SignedPauliOp
        return SignedPauliOp(self.array[key, :], validated=True)

    def __iter__(self):
        """
        Overloads the `in` operator to go through all SignedPauliOps.
        """
        return (
            SignedPauliOp(self.array[i, :], validated=True) for i in range(self.nstabs)
        )

    def __repr__(self) -> str:
        return f"StabArray({self.as_paulistrings})"


########################################################################################
# Methods of StabArray
########################################################################################


# pylint: disable=too-many-locals
def find_destabarray(  # pylint: disable=too-many-branches
    stabarr: StabArray, partial_destabarray: StabArray | None = None
) -> StabArray:
    """
    Given a stabilizer array of a state, find the destabilizer array.

    Parameters
    ----------
    stabilizer_array : StabArray
        The stabilizer array of a state. It must contain n independent stabilizers where
        n is the number of qubits.
    partial_destabarray : StabArray, optional
        A destabilizer array that partially destabilizes the state, i.e. it contains
        m < n operators that each anti-commute with exactly one state stabilizer. The
        final destabilizer array will contain these operators in the appropriate rows.

    Returns
    -------
    StabArray
        The destabilizer array.
    """
    if partial_destabarray is None:
        partial_destabarray = StabArray.trivial()

    nrows, ncols = stabarr.array.shape
    nqubs = stabarr.nqubits

    sarr_reduced = reduce_stabarray(stabarr)
    if sarr_reduced.nstabs != stabarr.nstabs:
        raise ValueError(
            "The StabArray should correspond to an irreducible set of stabilizers."
        )
    if stabarr.nstabs != nqubs:
        raise ValueError(
            "The StabArray should describe a full state, i.e. the number of stabilizers"
            f" ({stabarr.nstabs}) should be equal to the number of qubits ({nqubs})."
        )

    if partial_destabarray.nqubits != nqubs and not partial_destabarray.is_trivial:
        raise ValueError(
            "The partial destabilizer array must have the same number of qubits as the"
            " stabilizer array."
        )

    # Find for every operator of the partial destabilizer array the operators of the
    # stabilizer array that anti-commute with it.
    anti_commuting_indices = [
        [i for i, s_op in enumerate(stabarr) if pauliops_anti_commute(d_op, s_op)]
        for d_op in partial_destabarray
    ]

    if any(len(indices) != 1 for indices in anti_commuting_indices):
        raise ValueError(
            "Each operator of the partial destabilizer array must anti-commute with "
            "exactly one stabilizer."
        )

    # Make anti_commuting_indices a list of integers
    anti_commuting_indices = [indices[0] for indices in anti_commuting_indices]

    # Ensure that these indices are unique
    if len(anti_commuting_indices) != len(set(anti_commuting_indices)):
        raise ValueError(
            "Each operator of the partial destabilizer array must anti-commute with "
            "a different stabilizer."
        )

    # Find characteristic indices of stabilizer array
    char_idcs = [np.argmax(sarr_reduced.array[irow, :] == 1) for irow in range(nrows)]

    # Initialize d_array such that d_array and s_array
    # generate the full pauli set
    # Do not include the sign! --> (ncols-1)
    leftover_idx = set(range(ncols - 1)) - set(char_idcs)

    # builds an array which indicates which set of indices occur in leftover_idx
    d_array = np.zeros(stabarr.array.shape, dtype=SignedPauliOp.DTYPE)
    for row_idx, col_idx in enumerate(leftover_idx):
        d_array[row_idx, col_idx] = 1

    # translates column indices into the corresponding qubit indices
    row_qubit = np.array([col_idx % nqubs for col_idx in leftover_idx])

    # builds a dictionary out of qubits (keys) that appear twice and their
    # corresponding column indices (values)
    doubly_idxd_qubits = dict(
        filter(
            lambda tup: len(tup[1]) > 1,
            [(qubit, np.where(row_qubit == qubit)[0]) for qubit in set(row_qubit)],
        )
    )

    # Find which qubits have not been indexed by the d_array
    qubits_not_indexed = list(set(range(nqubs)) - set(row_qubit))

    # All possible anti-commuting single pauli combinations
    possible_combos = (
        ((1, 0), (0, 1)),  # X,Z
        ((1, 0), (1, 1)),  # X,Y
        ((0, 1), (1, 0)),  # Z,X
        ((0, 1), (1, 1)),  # Z,Y
        ((1, 1), (1, 0)),  # Y,X
        ((1, 1), (0, 1)),  # Y,Z
    )

    # Match each non-indexed qubit with the two rows of
    # a doubly indexed qubit.
    # Make sure that these rows eventually commute by using the
    # non-indexed qubit.
    # Simultaneously, ensure that this doesn't break the fact that
    # the concatenation of d_array,s_array generates the full pauli set.
    for non_indexed_qubit, (row1, row2) in zip(
        qubits_not_indexed, doubly_idxd_qubits.values(), strict=True
    ):
        # Cycle through the possible combinations
        # until we find one that doesn't break the
        # generation of the full pauli set
        for row1_vals, row2_vals in possible_combos:
            # assign the pauli of the the non-indexed qubits

            d_array[row1, non_indexed_qubit] = row1_vals[0]
            d_array[row1, non_indexed_qubit + nqubs] = row1_vals[1]

            d_array[row2, non_indexed_qubit] = row2_vals[0]
            d_array[row2, non_indexed_qubit + nqubs] = row2_vals[1]

            tableau_to_check = np.vstack((d_array, stabarr.array))
            if tableau_generates_pauli_group(tableau_to_check):
                # The set generates the pauli group.
                # We can continue to the next set of rows that
                # don't commute.
                break

    # Fix commutation and anti-commutation relations with the
    # stabilizer array.
    for i in range(nrows):
        s_operator = stabarr[i]

        # find with which d_array rows s_operator anti_commutes
        anti_commuting_rows = [
            j
            for j in range(nrows)
            if pauliops_anti_commute(s_operator, SignedPauliOp(d_array[j, :]))
        ]

        # Make sure that s_i anti-commutes
        # with its corresponding row
        if i not in anti_commuting_rows:
            # Get a row index that is larger than i
            # to not mess up progress.
            j_swap = next(j for j in anti_commuting_rows if j > i)

            # swap j_swap with i row
            d_array[[i, j_swap]] = d_array[[j_swap, i]]

            # Now that the row has been swapped into i,
            # remove it from the rows that need to be fixed
            anti_commuting_rows.remove(j_swap)
        else:
            # i is already in place.
            # Remove it from the rows that need to be fixed
            anti_commuting_rows.remove(i)

        # Multiply i into anti-commuting rows such that
        # s_i commutes with all d_j, j!=i
        for idx in anti_commuting_rows:
            d_array[idx, :] ^= d_array[i, :]

    # At this point, d_array is a valid destabilizer of the input StabArray.
    # Now we need to modify it to include the operators of partial_destabarray.

    # Set the rows of the destabilizer array from the partial destabilizer array
    for i, idx in enumerate(anti_commuting_indices):
        d_array[idx] = partial_destabarray.array[i]

    # Fix the anti-commutation relations between the destabilizer array rows occurred
    # due to setting the rows from the partial destabilizer array.
    if not partial_destabarray.is_trivial:
        # This if statement is to avoid unnecessary computations if the partial
        # destabilizer array is trivial.
        for i in set(range(nrows)) - set(anti_commuting_indices):
            # Rows with index i are going to be altered only.
            # This is why i does not include the anti_commuting_indices since they were
            # set and fixed in the previous step.
            for j in range(nrows):
                # When i == j, the operators trivially commute.
                if pauliops_anti_commute(
                    SignedPauliOp(d_array[i, :]), SignedPauliOp(d_array[j, :])
                ):
                    # Because the anti-commutation is not correct, we need to fix it.
                    # We do that by multiplying the j-th stabilizer into the i-th
                    # destabilizer.
                    # This works because [d_i, s_j] = 0 while {d_j, s_j} = 0.
                    d_array[i, :] ^= stabarr.array[j, :]

    return StabArray(d_array, validated=True)


def invert_bookkeeping_matrix(bookkeeping_matrix: np.ndarray) -> np.ndarray:
    """
    Invert the bookkeeping matrix.

    Parameters
    ----------
    bookkeeping_matrix : np.ndarray
        The bookkeeping matrix to be inverted.

    Returns
    -------
    np.ndarray
        The inverted bookkeeping matrix.
    """
    # We need to invert it and cast it to a boolean array
    return np.linalg.inv(bookkeeping_matrix).astype(bool)


def is_stabarray_equivalent(stab_arr0: StabArray, stab_arr1: StabArray) -> bool:
    """
    Check if two stabilizer arrays are equivalent.

    Parameters
    ----------
    stab_arr0 : np.ndarray
        A stabilizer array to be compared with another.
    stab_arr1 : np.ndarray
        A stabilizer array to be compared with another.

    Returns
    -------
    bool
        Whether the stabilizer arrays are equivalent.
    """
    if stab_arr0.nqubits != stab_arr1.nqubits:
        # the stabilizer arrays should have the same number of qubits to be equivalent
        return False

    stab_arr0_bge = reduce_stabarray(stab_arr0)
    stab_arr1_bge = reduce_stabarray(stab_arr1)
    return np.all(stab_arr0_bge.array == stab_arr1_bge.array)


def is_subset_of_stabarray(
    pauli_obj_sub: SignedPauliOp | StabArray, stab_arr_super: StabArray
) -> bool:
    """Check if a StabArray or SignedPauliOp is a subset of a StabArray.

    Parameters
    ----------
    pauli_obj_sub : SignedPauliOp | StabArray
        The SignedPauliOp or StabArray to be checked if it is a subset.
    stab_arr_super : StabArray
        The StabArray to be checked if it is a superset.

    Returns
    -------
    bool
        Whether the StabArray or SignedPauliOp is a subset of the StabArray.
    """
    # Convert the SignedPauliOp to a single row StabArray
    if isinstance(pauli_obj_sub, SignedPauliOp):
        pauli_obj_sub = StabArray.from_signed_pauli_ops([pauli_obj_sub], validated=True)

    # Check the validity of the inputs
    if not isinstance(pauli_obj_sub, StabArray):
        raise TypeError(
            "The pauli_obj_sub input should be a SignedPauliOp or a StabArray, not "
            f"{type(pauli_obj_sub)}."
        )
    if not isinstance(stab_arr_super, StabArray):
        raise TypeError(
            "The stab_arr_super input should be a StabArray, not "
            f"{type(stab_arr_super)}."
        )

    # If the number of qubits of the subarray is not the same as the superarray,
    # then the subarray cannot be a subset
    if pauli_obj_sub.nqubits != stab_arr_super.nqubits:
        return False

    # stack the arrays
    try:
        stab_arrays_merged = merge_stabarrays((stab_arr_super, pauli_obj_sub))
    except AntiCommutationError:
        # merging the arrays failed due to some operators anti-commuting
        return False
    # reduce the super set
    stab_arr_super_reduced = reduce_stabarray(stab_arr_super)
    # reduce the merged array
    stab_arrays_reduced = reduce_stabarray(stab_arrays_merged)
    # if the number of stabilizers of the superset is the same as the reduced,
    # then indeed stab_arr_sub was a subset
    return stab_arrays_reduced.nstabs == stab_arr_super_reduced.nstabs


def merge_stabarrays(stabarr_tuple: tuple[StabArray, ...]) -> StabArray:
    """
    Merges StabArrays.

    Parameters
    ----------
    stabarr_tuple : tuple[StabArray, ...]
        The tuple of StabArrays to be merged.

    Returns
    -------
    StabArray
        The merged StabArray.
    """
    if len(stabarr_tuple) < 2:
        raise ValueError("At least two StabArrays are needed to merge them.")

    # check if the StabArrays have the same number of qubits
    if any(stabarr.nqubits != stabarr_tuple[0].nqubits for stabarr in stabarr_tuple):
        raise ValueError("The StabArrays should have the same number of qubits.")

    for star1, star2 in combinations(stabarr_tuple, 2):
        # check if any of the operators of any pair anti-commute
        if any(pauliops_anti_commute(p1, p2) for p1 in star1 for p2 in star2):
            raise AntiCommutationError(
                "The StabArrays should only contain commuting operators."
            )

    return StabArray.from_signed_pauli_ops(
        [p for stabarr in stabarr_tuple for p in stabarr], validated=True
    )


def reduce_stabarray(stabarr: StabArray) -> StabArray:
    """
    Perform stabilizer BGE and then remove any trivial operators.

    Parameters
    ----------
    stabarr : StabArray
        The stabilizer array to be reduced.

    Returns
    -------
    StabArray
        The reduced stabilizer array.
    """
    stabarr_bge = stabarray_bge(stabarr)
    # remove empty stabilizers
    return StabArray.from_signed_pauli_ops(
        [p_op for p_op in stabarr_bge if not p_op.is_trivial],
        validated=True,
    )


def reduce_stabarray_with_bookkeeping(
    stabarr: StabArray,
) -> tuple[StabArray, np.ndarray]:
    """
    Perform stabilizer BGE and then remove any trivial operators. Also return the
    bookkeeping matrix, which is a matrix that keeps track of the row operations that
    were performed on the stabilizer array.

    Parameters
    ----------
    stabarr : StabArray
        The stabilizer array to be reduced.

    Returns
    -------
    tuple[StabArray, np.ndarray]
        A tuple containing the reduced stabilizer array and the bookkeeping matrix, i.e.
        the matrix that keeps track of the row operations that were performed on the
        stabilizer array.
    """
    stabarr_bge, bookkeeping_matrix = stabarray_bge_with_bookkeeping(stabarr)
    # find which operators are non trivial after BGE
    non_trivial_ops = [p_op for p_op in stabarr_bge if not p_op.is_trivial]
    return (
        StabArray.from_signed_pauli_ops(non_trivial_ops, validated=True),
        bookkeeping_matrix,
    )


def reindex_stabarray(stab_array: StabArray, new_idcs: list[int]) -> StabArray:
    """Returns an instance of a stabilizer array but with the qubit columns reindexed.

    Parameters
    ----------
    stab_array : StabArray
        The StabArray to be reindexed.
    new_idcs : list[int]
        The new indices.

    Returns
    -------
    StabArray
        The reindexed StabArray.

    Raises
    ------
    ValueError
        When the new_idcs are not unique or not as many as the n_qubits.
    """
    nqubits = stab_array.nqubits
    if len(new_idcs) != len(set(new_idcs)):
        raise ValueError("new_idcs has to have unique elements.")
    if len(new_idcs) != nqubits:
        raise ValueError("new_idcs should be as many as the n_qubits")

    array = stab_array.array.copy()
    reindexed_array = array.copy()

    new_idcs = np.array(new_idcs)

    # change the X's
    reindexed_array[:, 0:nqubits] = array[:, new_idcs]
    # change the Z's
    reindexed_array[:, nqubits : 2 * nqubits] = array[:, new_idcs + nqubits]

    return StabArray(reindexed_array, validated=True)


def swap_stabarray_rows(stabarr: StabArray, i: int, j: int) -> StabArray:
    """
    Swap two rows of a stabilizer array in-place.

    Parameters
    ----------
    stabarr : StabArray
        The stabilizer array to be modified.
    i : int
        The index of the first row.
    j : int
        The index of the second row.

    Returns
    -------
    StabArray
        The modified stabilizer array.
    """
    stabarr.array[[i, j]] = stabarr.array[[j, i]]
    return stabarr


def sparse_formatter(stab_list: set[str], **keywords) -> list[dict]:
    """Given a set of Pauli string operators,
    return a list of dicts, where dict keys are ``{'sign', 'X', 'Y', 'Z'}``. For keys
    ``X``, ``Y``, ``Z``, value is a list of indices/positions where the i-th qubit in
    the Pauli operator is respectively either X, Y, Z.

    There is a keyword option, to allow the user to pass a dictionary
    to convert *from* qubit index in CliffordSim, to some other indexing
    that is more convenient for the user.

    Parameters
    ----------
    stab_array : set[str]
        An unordered list of strings, where each string
        represents a Pauli string for a stabilizer operator in the set.

    keywords: {'convert_dict'}
        By supplying a dictionary that converts between qubit indices in CliffordSim
        into qubit IDs in an embedded QEC code, that user can return dicts where the
        values refer directly to IDs in an embedded QEC code.

    Returns
    -------
    output_format : list[dict]
        Reformats each string in the sparse format.
        E.g. +ZXIIYXZII gets returned as
        ``{'sign':'+', 'X':(1,5), 'Z':(0,6), 'Y':(4,)}``
    """

    output_format = []
    convert = False

    for key in keywords:
        if key == "convert_dict":
            assert isinstance(keywords["convert_dict"], dict)
            convert = True
            convert_dict = keywords["convert_dict"]

    for stab_op in stab_list:
        assert stab_op[0] == "+" or stab_op[0] == "-", "Improper format of Stab Op."
        formatted_dict = {"sign": stab_op[0], "X": [], "Y": [], "Z": []}

        xlist = [m.start() for m in re.finditer("X", stab_op[1:])]
        zlist = [m.start() for m in re.finditer("Z", stab_op[1:])]
        ylist = [m.start() for m in re.finditer("Y", stab_op[1:])]

        if len(xlist) != 0:
            if convert:
                xlist = [convert_dict[item] for item in xlist]
            formatted_dict["X"].extend(xlist)

        if len(zlist) != 0:
            if convert:
                zlist = [convert_dict[item] for item in zlist]
            formatted_dict["Z"].extend(zlist)

        if len(ylist) != 0:
            if convert:
                ylist = [convert_dict[item] for item in ylist]
            formatted_dict["Y"].extend(ylist)

        output_format.append(formatted_dict)

    return output_format


def stabarray_bge(stabarr: StabArray) -> StabArray:
    """
    Perform binary gaussian elimination on a stabilizer array.

    Parameters
    ----------
    stabarr : StabArray
        The stabilizer array to perform binary gaussian elimination on.

    Returns
    -------
    StabArray
        The stabilizer array in its reduced row echelon form.
    """
    return stabarray_bge_with_bookkeeping(stabarr)[0]


def stabarray_bge_with_bookkeeping(stabarr: StabArray) -> tuple[StabArray, np.ndarray]:
    """
    Performs binary gaussian elimination on a stabilizer array and returns
    the array in its reduced row echelon form.
    Reference:
    https://en.wikipedia.org/wiki/Row_echelon_form#Reduced_row_echelon_form

    Bookkeeping matrix explanation:
    It also returns the bookkeeping matrix, which is a matrix that keeps track of the
    row operations that were performed on the stabilizer array, i.e. how to transform
    the input stabilizer array into the reduced row echelon form. If the (i, j)-th
    element of the bookkeeping matrix is True, then the i-th stabilizer of the row
    echelon form contains the j-th row of the initial stabilizer array.
    In other words, to get the i-th stabilizer of the row echelon form, one should
    multiply all the stabilizers of the initial stabilizer array with index j that
    have a True value in the (i, j)-th element of the bookkeeping matrix.

    Parameters
    ----------
    stabarr : StabArray
        The stabilizer array to perform binary gaussian elimination on.

    Returns
    -------
    tuple[StabArray, np.ndarray]
        A tuple containing the reduced stabilizer array and the bookkeeping matrix, i.e.
        the matrix that keeps track of the row operations that were performed on the
        stabilizer array.
    """
    stabarr_copy = stabarr.copy()
    m, n = stabarr_copy.array.shape

    # Initialize the bookkeeping matrix
    bookkeeping_matrix = np.eye(m, dtype=bool)

    h = 0  # Initialization of the pivot row
    k = 0  # Initialization of the pivot column
    while h < m and k < n:
        # Find the k-th pivot:
        i_max = np.argmax(stabarr_copy.array[h:, k]) + h
        if stabarr_copy.array[i_max, k] == 0:
            # No pivot in this column, pass to next column
            k += 1
        else:
            # Swap rows h, imax
            stabarr_copy = swap_stabarray_rows(stabarr_copy, h, i_max)

            # Update the bookkeeping matrix
            bookkeeping_matrix[[h, i_max]] = bookkeeping_matrix[[i_max, h]]

            # Need to apply rowsum to eliminate 1s
            # on column k.

            # Do for all rows excluding pivot
            irange = np.concatenate([np.arange(h), np.arange(h + 1, m)])
            # Find rows with 1 in column k
            irange_with_1_in_col_k = irange[stabarr_copy.array[irange, k] == 1]

            # Apply rowsum to row idx if if st_ar[idx, k] == 1
            stabarr_copy = reduce(
                lambda st, i: rowsum(st, i, h),
                irange_with_1_in_col_k,
                stabarr_copy,
            )

            # XOR h into the irange_with_1_in_col_k in the bookkeeping_matrix
            bookkeeping_matrix[irange_with_1_in_col_k] = np.bitwise_xor(
                bookkeeping_matrix[irange_with_1_in_col_k], bookkeeping_matrix[h]
            )

            # Increase pivot row and column
            h += 1
            k += 1

    return stabarr_copy, bookkeeping_matrix


def stabarray_standard_form(stabarr: StabArray) -> tuple[StabArray, int, list[int]]:
    """Transform a stabilizer array into a standard form. Note that this means that the
    stabilizer array is reduced and reindexed.

        1. Reduce the stabilizer array.
        
        2. Reorder the qubits such that the X part of the stabilizer array becomes of 
        the form
        
            (I, A)
            (0, 0)

        where I is the identity matrix of size r and A is a r x (n-r) matrix and the 0 \
        arrays have n-k-r rows.
        
        3. Multiply rows of the stabilizer array so that it becomes of the form
        
            (I, A1, A2 | B, 0, C)
            (0,  0,  0 | D, I, E)
            
        where I is the identity matrix, A1, A2, B, C, D and E are matrices.
    
    Dimension explanation:
    
        - The left blocks comprise the X part of the stabilizer array and the right \
        blocks comprise the Z part of the stabilizer array.
        - The top blocks have r rows and the bottom blocks have n-k-r rows, where n is \
        the number of qubits, k is the number of logical qubits and r is the rank of \
        the X part of the stabilizer array.
        - From left to right, the number of columns in the blocks are r, n-k-r, k for \
        each of the Z and X parts.
        
    More info:
    
        Nielsen, M. A., & Chuang, I. L. (2011). (p.470-471) Quantum Computation and 
        Quantum Information: 10th Anniversary Edition. Cambridge University Press

    Parameters
    ----------
    stabarr : StabArray
        The stabilizer array to be transformed.

    Returns
    -------
    tuple[StabArray, int, list[int]]
        A tuple containing the reindexed stabilizer array, the rank of the X part of the
        stabilizer array and the inverse qubit map. The inverse qubit map can be used to
        reindex the stabilizer array to the original qubit order.
    """
    # Reduce the stabilizer array
    stabarr_reduced = reduce_stabarray(stabarr)

    # Name quantities to shorten notation
    n = stabarr.nqubits  # Number of qubits
    k = stabarr.nqubits - stabarr_reduced.nstabs  # Number of logical qubits

    # Find the characteristic indices of the stabilizer array, i.e. the indices of the
    # first non-zero entry in each row
    characteristic_idx = [
        np.argmax(signed_pauli_op.array) for signed_pauli_op in stabarr_reduced
    ]
    # Find the indices of the characteristic rows that correspond to the X part of the
    # stabilizer array
    characteristic_idx_x = [idx for idx in characteristic_idx if idx < n]

    r = len(characteristic_idx_x)  # The rank of the X part of the stabilizer array

    # Find the indices of the rows that are not characteristic of the X part
    rest_idx = [idx for idx in range(n) if idx not in characteristic_idx_x]

    # We need to order the rest_idx such that the identity array can be generated
    # at the bottom of the stabilizer array. This is a concatenation of the I, E blocks
    # in the standard form of the stabilizer array.
    # 1. Isolate the bottom right block of the stabilizer array
    bottom_right_ie_block = stabarr_reduced.z[r:, np.array(rest_idx)]
    # 2. Perform Binary Gaussian Elimination
    bottom_right_ie_block_bge = binary_gaussian_elimination(bottom_right_ie_block)
    # 3. Find the characteristic indices of the bottom right block
    ie_block_characteristic_idx = [np.argmax(row) for row in bottom_right_ie_block_bge]
    # 4. Find the indices of the rows that are not characteristic
    ie_block_non_characteristic_idx = [
        idx for idx in range(len(rest_idx)) if idx not in ie_block_characteristic_idx
    ]
    # 5. Reorder the rest_idx such that the characteristic indices come first. This
    # means that the identity matrix can be generated at the bottom of the stabilizer
    rest_idx_reordered = [
        rest_idx[idx]
        for idx in ie_block_characteristic_idx + ie_block_non_characteristic_idx
    ]

    # Construct a new map that reorders the qubits such that the X part of the
    # stabilizer array becomes of the form:
    # (I, A)
    # (0, 0)
    # where I is the identity matrix of size r and A is a r x (n-r) matrix and the 0
    # arrays have n-k-r columns
    qbit_map = characteristic_idx_x + rest_idx_reordered
    inv_qbit_map = [qbit_map.index(idx) for idx in range(stabarr.nqubits)]

    # Reindex the stabilizer array according to the new map.
    # Only the qubits are reindexed by swapping columns in the X and Z parts of the
    # stabilizer array.
    reindexed_stabarr = reindex_stabarray(stabarr_reduced, qbit_map)

    # Create the identity matrix on the bottom part of the reindexed stabilizer array
    for i in range(n - k - r):
        # Find the array that is to be made an identity matrix
        arr_to_be_eye = reindexed_stabarr.z[r:, r : n - k]
        # NOTE: Because arr_to_be_eye is a subarray of the reindexed_stabarr, we may
        # need to offset the indices by r when modifying the reindexed_stabarr

        # if the i-th diagonal element is not 1, then swap the i-th row with a row that
        # has 1 in the i-th column
        if arr_to_be_eye[i, i] == 0:
            # find a row with 1 in the i-th column and swap it with the i-th row
            row_with_one = np.argmax(arr_to_be_eye[i:, i] == 1) + i
            # if such a row is not found, then raise an error
            if row_with_one == i:
                raise ValueError(
                    "Cannot create identity matrix inside the reindexed "
                    "stabilizer array"
                )
                # NOTE: if this ever fails for a valid code, it may be worth checking if
                # the error can be resolved by reindexing the StabArray further

            # swap the rows
            reindexed_stabarr = swap_stabarray_rows(
                reindexed_stabarr, r + i, r + row_with_one
            )

        # make the i-th column of the reindexed stabilizer array to be that of an
        # identity matrix
        rows_to_make_zero = [
            j for j in range(n - k - r) if j != i and arr_to_be_eye[j, i] == 1
        ]
        reindexed_stabarr = reduce(
            lambda acc, row, _i=i: rowsum(acc, r + row, r + _i),
            rows_to_make_zero,
            reindexed_stabarr,
        )

    return reindexed_stabarr, r, inv_qbit_map


def subtract_stabarrays(
    stab_array: StabArray, stab_array_to_remove: StabArray
) -> StabArray:
    """Given 2 stabilizer arrays, remove from the first one any
    occurrence of stabilizers found in the second one.

    Parameters
    ----------
    stab_array : StabArray
        The StabArray from which the other is subtracted.
    stab_array_to_remove : StabArray
        The StabArray that is subtracted from the other.

    Returns
    -------
    StabArray
        The resulting StabArray.
    """
    # check if the StabArrays have the same number of qubits
    if stab_array.nqubits != stab_array_to_remove.nqubits:
        raise ValueError("The StabArrays should have the same number of qubits.")

    if any(
        pauliops_anti_commute(p, q) for p in stab_array_to_remove for q in stab_array
    ):
        raise ValueError("The operators of the StabArrays should not anti-commute.")

    # set stab_array in its reduced form
    stab_array_to_remove = reduce_stabarray(stab_array_to_remove)

    # Do the subtraction
    subtracted_paulis = []
    for pauli_op in stab_array:
        # Create a temporary StabArray with the operator from which we will subtract
        # the operators of the other StabArray
        # This gives us access of the rowsum function
        temp_stabarr = StabArray.from_signed_pauli_ops(
            list(stab_array_to_remove) + [pauli_op],
            validated=True,
        )

        # Find for every operator A in stab_array_to_remove the characteristic index:
        # np.argmax(st.array[idx, :] == 1)
        # If at the characteristic index the last operator has 1, then multiply
        # A into the last operator.
        temp_stabarr = reduce(
            lambda st, idx, _pauli=pauli_op: (
                rowsum(st, -1, idx)
                if _pauli.array[np.argmax(st.array[idx, :] == 1)] == 1
                else st
            ),
            range(temp_stabarr.nstabs - 1),
            temp_stabarr,
        )

        subtracted_paulis.append(temp_stabarr[-1])

    # Combine the result
    result_arr = StabArray.from_signed_pauli_ops(subtracted_paulis, validated=True)

    # Reduce the result
    return reduce_stabarray(result_arr)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Circuit
from loom.eka.operations.code_operation import (
    LogicalX,
    LogicalZ,
    LogicalY,
)

from ..interpretation_step import InterpretationStep


def logical_pauli(
    interpretation_step: InterpretationStep,
    operation: LogicalX | LogicalY | LogicalZ,
    same_timeslice: bool,
    debug_mode: bool,  # pylint: disable=unused-argument
) -> InterpretationStep:
    """
    Apply a logical X, Y or Z operator.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step to which the operation should be applied.
    operation : LogicalX | LogicalY | LogicalZ
        The operation to be applied, can either be a logical X, Y or Z operation.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        New InterpretationStep containing all modifications due to the logical pauli
        gate.
    """
    # Note: This method is used for LogicalX, LogicalY and LogicalZ operations.
    # The operation is applied to the interpretation step by appending the
    # corresponding circuit to the interpretation step.
    # No other changes are made to the interpretation step.

    block = interpretation_step.get_block(operation.input_block_name)
    logical_qubit = operation.logical_qubit

    # Check if the logical qubit exists
    if logical_qubit >= block.n_logical_qubits:
        raise ValueError(
            f"Logical qubit {logical_qubit} does not exist in block "
            f"{block.unique_label}"
        )

    # Check which logical operator to apply and get the corresponding Pauli
    # operator(s)
    match operation.__class__.__name__:
        case "LogicalX":
            logical_operators = [block.logical_x_operators[logical_qubit]]
        case "LogicalZ":
            logical_operators = [block.logical_z_operators[logical_qubit]]
        case "LogicalY":
            logical_operators = [
                block.logical_x_operators[logical_qubit],
                block.logical_z_operators[logical_qubit],
            ]
        case _:
            raise ValueError(f"Operation {operation.__class__.__name__} not supported")

    # Create the circuit
    logical_operation_circuit = Circuit(
        name=(
            f"{operation.__class__.__name__} on block {block.unique_label}, "
            f"logical qubit {logical_qubit}"
        ),
        circuit=[
            [
                Circuit(pauli, channels=[interpretation_step.get_channel_MUT(qb)])
                for qb, pauli in zip(
                    logical_operator.data_qubits, logical_operator.pauli, strict=True
                )
            ]
            for logical_operator in logical_operators
        ],
    )

    # Append the circuit
    interpretation_step.append_circuit_MUT(logical_operation_circuit, same_timeslice)

    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=duplicate-code
from dataclasses import dataclass, field
from enum import Enum
from typing import Optional

from .base_operation import Operation, OpType
from .controlled_operation import has_ccontrol


@dataclass
@has_ccontrol
class ClassicalOperation(Operation):
    """
    Operations of this type manipulate classical bits within the Engine during runtime.
    """

    operation_type: Enum = field(default=OpType.CLASSICAL, init=False)


@dataclass
class ClassicalNOT(ClassicalOperation):
    """Apply a classical NOT operation on a bit.

    Parameters
    ----------
    reg_name: str
        The name of the classical register whose bit to apply the Operation to.
    bit_order: int
        The numerical order of the bit within the classical register.
    bit_id: str
        The ID of the bit within the classical register.
    """

    name: str = field(default="ClassicalNOT", init=False)
    reg_name: str
    bit_order: int = field(default=None)
    bit_id: str = field(default=None)

    def __post_init__(self):
        if self.bit_order is None and self.bit_id is None:
            raise ValueError(
                "Either the bit order or the bit id of the classical bit has to be "
                "specified."
            )
        if self.bit_order is not None and self.bit_id is not None:
            raise ValueError(
                "Both bit_order and bit_id cannot be specified together. Only input 1 "
                "parameter, not both."
            )


@dataclass
class ClassicalTwoBitOperation(ClassicalOperation):
    """Classical 2 Bit Operations that require 2 bits as input and 1 bit as output.

    Parameters
    ----------
    reg_name: str
        The name of the classical register whose bits is used as input to evaluate the
        2-bit operation.
    input_bit_order: list[int]
        The order of the bits within the classical register that will be used as inputs
        for the 2-bit operation.
    input_bit_ids: list[str]
        The bit IDs of the bits within the classical reigster that will be used as
        inputs for the 2-bit operation.
    output_reg_name: Optional[str]
        The name of the classical register where the output of the 2-bit operation will
        be written to. If not provided, this is equal to the register referred to by
        reg_name.
    write_bit_order: int
        The order of the bit within the classical register, specified by
        output_reg_name, where the output of the 2-bit operation will be written to.
    write_bit_ids: str
        The bit ID of the bit within the classical register, specified by
        output_reg_name, where the output of the 2-bit operation will be written to.
    """

    reg_name: str
    input_bit_order: list[int] = field(default_factory=list)
    input_bit_ids: list[str] = field(default_factory=list)
    output_reg_name: Optional[str] = field(default=None)
    write_bit_order: int = field(default=None)
    write_bit_id: str = field(default=None)

    def __post_init__(self):
        if (bool(self.input_bit_order)) ^ (bool(self.input_bit_ids)):
            if not (len(self.input_bit_order) == 2) ^ (len(self.input_bit_ids) == 2):
                raise ValueError(
                    "There must be at least 2 bits referenced using their bit ordering "
                    "or their bit IDs for this operation."
                )
        else:
            raise ValueError(
                "Both input_bit_order and input_bit_ids cannot be specified together."
                "Only input 1 parameter, not both."
            )

        if not (self.write_bit_order is not None) ^ (self.write_bit_id is not None):
            raise ValueError(
                "Both input_bit_order and input_bit_ids cannot be specified together."
                "Only input 1 parameter, not both."
            )

        if self.output_reg_name is None:
            self.output_reg_name = self.reg_name


@dataclass
class ClassicalOR(ClassicalTwoBitOperation):
    """Evaluates a classical OR operation from 2 input bits and writes the output onto
    a referenced bit, known as the write bit.
    """

    name: str = field(default="ClassicalOR", init=False)


@dataclass
class ClassicalAND(ClassicalTwoBitOperation):
    """
    Evaluates a classical AND operation from 2 input bits and writes the output onto
    a referenced bit, known as the write bit.
    """

    name: str = field(default="ClassicalAND", init=False)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from typing import List, Dict
import traceback

from .tableau import Tableau
from .pauli_frame import PauliFrame
from .data_store import DataStore
from .moments.base_moment import Moment


class Invoker:
    """
    The main role of the Invoker is to handle the interactions between an input
    Moment, the Tableau, PauliFrames and DataStore.
    """

    def __init__(
        self,
        input_tableau: Tableau,
        pauli_frames_forward: List[PauliFrame],
        pauli_frames_backward: List[PauliFrame],
        data_store: DataStore,
        registry: Dict,
    ):
        self.input_tableau = input_tableau
        self.pauli_frames_forward = pauli_frames_forward
        self.pauli_frames_backward = pauli_frames_backward
        self.data_store = data_store
        self.registry = registry

    def transform_tab(self, input_moment: Moment) -> bool:
        """
        Transform the input_tableau according to the input moment.
        The data_store and registry are also passed to the moment's transform_tab
        method to record any potential output measurement and allow for classical
        control.
        """
        try:
            self.data_store.set_time_step(input_moment.time_step)
            input_moment.transform_tab(
                self.input_tableau, self.data_store, registry=self.registry
            )
            return True
        except Exception:  # pylint: disable=broad-exception-caught
            print(traceback.format_exc())
            return False

    def transform_pf(self, input_moment: Moment) -> bool:
        """
        Transform the pauli frames according to the input moment.
        The data_store is also passed to the moment's transform_pf method to record any
        potential output pauli frame state.
        """
        try:
            self.data_store.set_time_step(input_moment.time_step)
            # Apply transformation to all pauli frames
            input_moment.transform_pf(self.pauli_frames_forward, self.data_store)
            return True
        except Exception:  # pylint: disable=broad-exception-caught
            print(traceback.format_exc())
            return False

    def transform_pf_back(self, input_moment: Moment) -> bool:
        """
        Transform the pauli frames backwards according to the input moment.
        The data_store is also passed to the moment's transform_pf_back method to
        record any potential output pauli frame state.
        """
        try:
            self.data_store.set_time_step(input_moment.time_step)
            # Apply transformation to all pauli frames
            input_moment.transform_pf_back(self.pauli_frames_backward, self.data_store)
            return True
        except Exception:  # pylint: disable=broad-exception-caught
            print(traceback.format_exc())
            return False



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=duplicate-code
from itertools import product

from loom.eka import Circuit, ChannelType, PauliOperator, Stabilizer
from loom.eka.operations import Shrink
from loom.eka.utilities import Orientation
from loom.interpreter import InterpretationStep
from loom.interpreter.utilities import Cbit
from loom.interpreter.applicator import generate_detectors, generate_syndromes

from ..code_factory import RotatedSurfaceCode


def shrink_consistency_checks(
    interpretation_step: InterpretationStep,
    operation: Shrink,
) -> RotatedSurfaceCode:
    """
    Check if the Shrink operation is valid for the given state of blocks

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block to shrink.
    operation : Shrink
        Shrink operation description.

    Returns
    -------
    RotatedSurfaceCode
        The block to be shrunk.
    """
    block = interpretation_step.get_block(operation.input_block_name)

    # Check block type
    if not isinstance(block, RotatedSurfaceCode):
        raise TypeError(
            f"The shrink operation is not supported for {type(block)} blocks."
        )

    is_horizontal = (
        Orientation.from_direction(operation.direction) == Orientation.HORIZONTAL
    )
    block_distance = block.size[0] if is_horizontal else block.size[1]

    # Limit the possible shrink lengths (between 1 and block_distance-2)
    if operation.length < 1 or operation.length > block_distance - 2:
        raise ValueError(
            f"Shrink length {operation.length} is not valid. "
            f"Must be between 1 and {block_distance - 2} for the selected block."
        )

    return block


# pylint: disable=too-many-statements, too-many-locals, too-many-branches
def shrink(
    interpretation_step: InterpretationStep,
    operation: Shrink,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Shrink a Block in the specified direction.

    The algorithm is the following:

    - A.) DATA QUBITS
    
        - A.1) Find data qubits which are measured in the shrink
        
    - B.) CIRCUIT
    
        - B.1) Create classical channels for all data qubit measurements
        - B.2) Create a measurement circuit for every measured data qubit
        - B.3) Append the measurement circuits to the InterpretationStep circuit. \
        If needed, apply a basis change
            
    - C.) STABILIZERS
    
        - C.1) Find stabilizers which are completely removed
        - C.2) Find stabilizers which have to be reduced in weight
        - C.3) Create new stabilizers with reduced weight
        - C.4) Update ``stabilizer_evolution`` and ``stabilizer_updates`` for the \
        stabilizers which have been reduced in weight
        - C.5) Combine the stabilizers to get a new set of stabilizers
        - C.6) Create syndromes for fully measured stabilizers
        - C.7) Create detectors for fully measured stabilizers
        
    - D.) LOGICAL OPERATORS
    
        - D.1) Find logical operators which do not have to be modified
        
    - E.) NEW BLOCK AND NEW INTERPRETATION STEP
    
        - E.1) Update the stabilizer to circuit mapping
        - E.2) Create the new block
        - E.3) Update the block history

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block to shrink.
    operation : Shrink
        Shrink operation description.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block

    Returns
    -------
    InterpretationStep
        Interpretation step after the shrink operation.
    """

    block = shrink_consistency_checks(
        interpretation_step=interpretation_step, operation=operation
    )
    boundary_type = block.boundary_type(operation.direction)

    # A) DATA QUBITS

    # Generate a list of vectors by which the boundary qubits should be shifted to
    # obtain all data qubits that should be measured
    match operation.direction:
        case "left":
            shift_vectors = [(i, 0, 0) for i in range(operation.length)]
        case "right":
            shift_vectors = [(-i, 0, 0) for i in range(operation.length)]
        case "top":
            shift_vectors = [(0, i, 0) for i in range(operation.length)]
        case "bottom":
            shift_vectors = [(0, -i, 0) for i in range(operation.length)]

    # Get a list of all data qubits to measure
    combinations = product(
        block.boundary_qubits(direction=operation.direction), shift_vectors
    )
    qubits_to_measure = [
        tuple(coord1 + coord2 for coord1, coord2 in zip(qubit, shift_vect, strict=True))
        for qubit, shift_vect in combinations
    ]

    # B) CIRCUIT
    # B.1) Create classical channels for all data qubit measurements
    # So far, the classical channels are all named "c_{q}" for some qubit q.
    cbits = [interpretation_step.get_new_cbit_MUT(f"c_{q}") for q in qubits_to_measure]
    cbit_channels = [
        interpretation_step.get_channel_MUT(
            f"{cbit[0]}_{cbit[1]}", channel_type=ChannelType.CLASSICAL
        )
        for cbit in cbits
    ]
    # B.2) Create a measurement circuit for every measured data qubit and create
    # the sequence of gates
    measure_circuit_seq = [
        [
            Circuit(
                "Measurement",
                channels=[interpretation_step.get_channel_MUT(qb), cbit_channels[i]],
            )
            for i, qb in enumerate(qubits_to_measure)
        ]
    ]
    # B.3) Append the measurement circuits to the InterpretationStep circuit
    # If needed, apply a basis change before the measurement to measure in the right
    # basis
    if boundary_type == "X":
        # If the boundary type is X, the data qubit can directly be read out in the
        # Z basis
        shrink_circuit_list = measure_circuit_seq
    else:
        # If the boundary type is Z, the 2-bodies stabilizers are X, the data qubits
        # have to be read out in the X basis. Apply Hadamard gates before the
        # Z measurement to effectively measure in the X basis
        basis_change_circuit_seq = [
            [
                Circuit("H", channels=[interpretation_step.get_channel_MUT(qb)])
                for qb in qubits_to_measure
            ]
        ]
        # We add two sequences of gates to the circuit, i.e. 2 timesteps
        shrink_circuit_list = basis_change_circuit_seq + measure_circuit_seq
    # Construct the final circuit and append it to InterpretationStep.circuit
    shrink_circuit = Circuit(
        name=(
            f"Shrink {block.unique_label} by {operation.length} "
            f"from {operation.direction}"
        ),
        circuit=shrink_circuit_list,
    )
    interpretation_step.append_circuit_MUT(shrink_circuit, same_timeslice)

    # C) STABILIZERS

    # C.1) Find stabilizers which are completely removed
    stabs_to_remove = [
        stab
        for stab in block.stabilizers
        if any(qb in qubits_to_measure for qb in stab.data_qubits)
    ]

    # C.2) Find stabilizers which have to be reduced in weight
    old_stabs_to_reduce_weight = [
        stab
        for stab in block.stabilizers
        if stab.pauli[0] != boundary_type
        and len([qb for qb in stab.data_qubits if qb in qubits_to_measure]) == 2
        and len(stab.data_qubits) == 4
    ]

    # C.3) Create new stabilizers with reduced weight
    new_stabs_reduced_weight = [
        Stabilizer(
            pauli="".join(
                stab.pauli[i]
                for i, qb in enumerate(stab.data_qubits)
                if qb not in qubits_to_measure
            ),
            data_qubits=[qb for qb in stab.data_qubits if qb not in qubits_to_measure],
            ancilla_qubits=stab.ancilla_qubits,
        )
        for stab in old_stabs_to_reduce_weight
    ]
    stab_map_weight4_to_weight2 = {
        new_stabs_reduced_weight[i].uuid: (stab.uuid,)
        for i, stab in enumerate(old_stabs_to_reduce_weight)
    }
    # C.4) Update `stabilizer_evolution` and `stabilizer_updates` for the
    #      stabilizers which have been reduced in weight
    interpretation_step.stabilizer_evolution.update(stab_map_weight4_to_weight2)
    # Stabilizer updates: Take data qubit measurements into account
    for i, stab in enumerate(new_stabs_reduced_weight):
        previous_updates = (
            interpretation_step.stabilizer_updates[stab.uuid]
            if stab.uuid in interpretation_step.stabilizer_updates.keys()
            else ()
        )
        old_stab = old_stabs_to_reduce_weight[i]
        # Find the data qubits of this stabilizer which are measured and whose
        # corresponding cbit has to be included in the stabilizer update
        qbs_measured = [qb for qb in old_stab.data_qubits if qb in qubits_to_measure]
        # Find the cbits and include them in `stabilizer_updates`
        cbit_indices = [qubits_to_measure.index(qb) for qb in qbs_measured]
        new_updates = tuple(cbits[cbit_idx] for cbit_idx in cbit_indices)
        interpretation_step.stabilizer_updates[stab.uuid] = (
            previous_updates + new_updates
        )

    # C.5) Combine the stabilizers to get the new set of stabilizers
    new_stabs = set(block.stabilizers) - set(stabs_to_remove) | set(
        new_stabs_reduced_weight
    )

    # C.6) Create syndromes for fully measured stabilizers (in the right basis)
    fully_measured_stabs = [
        stab
        for stab in block.stabilizers
        if (
            all(qb in qubits_to_measure for qb in stab.data_qubits)
            and set(stab.pauli) == ({"Z"} if boundary_type == "X" else {"X"})
        )
    ]
    # Create the syndromes fo fully measuerd stabilizers, we need to order the
    # measurements in an order matching the fully_measured_stabs
    syndromes = generate_syndromes(
        interpretation_step,
        stabilizers=fully_measured_stabs,
        block=block,
        stab_measurements=tuple(
            tuple(
                cbit
                for q, cbit in product(stab.data_qubits, cbits)
                if cbit[0].split("_")[1] == str(q)
            )  # Match the order of cbits and stabilizers
            for stab in fully_measured_stabs
        ),
    )

    # C.7) Create the detectors for the fully measured stabilizers
    detectors = generate_detectors(interpretation_step, syndromes)
    # Finally append both syndromes and detectors
    interpretation_step.append_syndromes_MUT(syndromes)
    interpretation_step.append_detectors_MUT(detectors)

    # D) LOGICAL OPERATORS
    # D.1) Find logical operators which do not have to be modified

    def shorten_log_op(
        old_op: PauliOperator, cbits: list[Cbit]
    ) -> tuple[PauliOperator, dict]:
        """
        Return a shortened logical operator where the measured data qubits are
        removed. Note that this only works if the data qubits are measured in the
        right basis. Also, a new logical operator update dict is returned containing
        the new cbit measurements for the respective operator.

        Parameters
        ----------
        old_op : PauliOperator
            Old logical X/Z operator which has to be shortened
        cbits : list[Cbit]
            Measurements tied to the logical X/Z operator

        Returns
        -------
        PauliOperator
            Shortened logical X/Z operator
        tuple[Cbit, ...]
            Measurements tied to the short logical X/Z operator
        """
        # New shorter operator
        new_op = PauliOperator(
            pauli="".join(
                old_op.pauli[i]
                for i, qb in enumerate(old_op.data_qubits)
                if qb not in qubits_to_measure
            ),
            data_qubits=[
                qb for qb in old_op.data_qubits if qb not in qubits_to_measure
            ],
            uuid=old_op.uuid,
        )

        # Logical operator updates: Include the new cbit measurements
        cbit_indices = [
            qubits_to_measure.index(qb)
            for qb in old_op.data_qubits
            if qb in qubits_to_measure
        ]
        measurements = tuple(cbits[idx] for idx in cbit_indices)

        return new_op, measurements

    def get_new_log_op(
        int_step: InterpretationStep,
        old_op: PauliOperator,
        new_upper_left_corner: tuple[int, int, int],
        cbits: list[Cbit],
    ) -> tuple[PauliOperator, tuple[Cbit, ...], dict]:
        """
        Return new `PauliOperator` for a logical operator which has to be
        modified due to the shrink. Also, a new logical operator update dict is
        returned containing the cbit measurements for the respective operator and a
        new logical operator evolution dict is returned containing the mapping from
        the new operator to the old operator and eventual necessary stabilizers.

        Parameters
        ----------
        int_step : InterpretationStep
            Interpretation step containing the block to shrink
        old_op : PauliOperator
            Old logical X/Z operator to be changed
        new_upper_left_corner : tuple[int, int, int]
            New upper-left corner of the block after the shrink

        Returns
        -------
        PauliOperator
            New logical X/Z operator
        tuple[Cbit, ...]
            Measurements tied to the new logical X/Z operator
        dict
            Updated logical X/Z operator evolution dictionary
        """
        # Check in which bases the measured data qubits are included in the logical
        # operator
        measured_paulis = set(
            old_op.pauli[i]
            for i, qb in enumerate(old_op.data_qubits)
            if qb in qubits_to_measure
        )
        if len(measured_paulis) > 1:
            raise RuntimeError(
                f"Cannot update logical operator {old_op} during the shrink "
                "because it contains multiple different Paulis inside the region "
                "of data qubits that are measured in the shrink."
            )
        if measured_paulis == {"Y"}:
            raise RuntimeError(
                f"Cannot update logical operator {old_op} during the shrink "
                "because it contains a Y Pauli inside the region of data qubits "
                "that are measured in the shrink."
            )
        if list(measured_paulis)[0] != boundary_type:
            # This is the case where the data qubits are measured in the "right"
            # basis, i.e. in the same basis in which they are included in the
            # logical operator. In this case, the logical operator is simply shrunk
            # and those data qubits are added to the update list for later
            # processing
            new_op, measurements = shorten_log_op(old_op, cbits)
            new_log_op_evolution = {new_op.uuid: (old_op.uuid,)}
            return new_op, measurements, new_log_op_evolution
        if list(measured_paulis)[0] == boundary_type:
            # This is the case where the data qubits are measured in a basis which
            # is different to the basis in which they are included in the logical
            # operator.  In this case, the logical operator has to be moved into the
            # region of the block that is not measured. This can be done by
            # multiplying with stabilizers
            new_op, stabs_required = block.get_shifted_equivalent_logical_operator(
                old_op, new_upper_left_corner
            )
            # Add to the updates the cbits associated with the measured qubits and the
            # stabilizers required for the shift.
            cbits_required = int_step.retrieve_cbits_from_stabilizers(
                stabs_required, block
            )
            new_log_op_evolution = {
                new_op.uuid: (old_op.uuid,)
                + tuple(stab.uuid for stab in stabs_required)
            }
            return new_op, cbits_required, new_log_op_evolution

        raise RuntimeError("This should not happen. Please check the code for errors.")

    # Those operators which have at least one qubit inside the measured region
    # have to be updated
    new_qubits = [q for stab in new_stabs for q in stab.data_qubits]
    new_upleft_qubit = min(new_qubits, key=lambda x: x[0] + x[1])
    new_log_x_ops, new_log_z_ops = [], []
    for op in block.logical_x_operators:
        if any(qb in qubits_to_measure for qb in op.data_qubits):
            new_op, measurements, new_evolution = get_new_log_op(
                interpretation_step,
                op,
                new_upleft_qubit,
                cbits,
            )
            new_log_x_ops.append(new_op)
            interpretation_step.logical_x_evolution.update(new_evolution)
            interpretation_step.update_logical_operator_updates_MUT(
                operator_type="X",
                logical_operator_id=new_op.uuid,
                new_updates=measurements,
                inherit_updates=(new_op.uuid != op.uuid),
            )
        else:
            new_log_x_ops.append(op)
    for op in block.logical_z_operators:
        if any(qb in qubits_to_measure for qb in op.data_qubits):
            new_op, measurements, new_evolution = get_new_log_op(
                interpretation_step,
                op,
                new_upleft_qubit,
                cbits,
            )
            new_log_z_ops.append(new_op)
            interpretation_step.logical_z_evolution.update(new_evolution)
            interpretation_step.update_logical_operator_updates_MUT(
                operator_type="Z",
                logical_operator_id=new_op.uuid,
                new_updates=measurements,
                inherit_updates=(new_op.uuid != op.uuid),
            )
        else:
            new_log_z_ops.append(op)

    # E) NEW BLOCK
    # E.1) Update the stabilizer to circuit mapping
    new_stabilizer_to_circuit = block.stabilizer_to_circuit
    # Remove those stabilizers which do not exist anymore
    for stab in stabs_to_remove:
        new_stabilizer_to_circuit.pop(stab.uuid)
    # Associate the new stabilizers with the right syndrome circuits
    for stab in new_stabs_reduced_weight:
        # The default creation in the RotatedSurfaceCode class creates
        # SyndromeCircuits with the name "{direction}-{pauli}" where direction is
        # left, right, top or bottom. Find the SyndromeCircuit for this name and
        # take its uuid
        circ_name = f"{operation.direction}-{stab.pauli.lower()}"
        new_stabilizer_to_circuit[stab.uuid] = [
            circ for circ in block.syndrome_circuits if circ.name == circ_name
        ][0].uuid

    # E.2) Create the new block
    new_block = RotatedSurfaceCode(
        stabilizers=list(new_stabs),
        logical_x_operators=new_log_x_ops,
        logical_z_operators=new_log_z_ops,
        unique_label=block.unique_label,
        syndrome_circuits=block.syndrome_circuits,
        stabilizer_to_circuit=new_stabilizer_to_circuit,
        skip_validation=not debug_mode,
    )

    # E.3) Update the block history
    interpretation_step.update_block_history_and_evolution_MUT(
        new_blocks=(new_block,),
        old_blocks=(block,),
    )

    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import ast
from collections import defaultdict
from dataclasses import dataclass
from functools import cached_property, partial
from typing import Any, Callable
import traceback

from numpy.typing import NDArray
import numpy as np
from pydantic import Field
import stim


from ..eka import Circuit, Channel, IfElseCircuit
from ..interpreter import InterpretationStep

from .op_signature import (
    CLIFFORD_GATES_SIGNATURE,
    OpSignature,
    OpType,
)
from .converter import Converter
from .circuit_error_model import CircuitErrorModel, ErrorType, ApplicationMode

# pylint: disable=no-member

# Type of the output of the Convert function.
# stim.Circuit and dictionaries mapping classical channel to stim measurement record
# indices and quantum channel to stim qubit coordinate and target.
StimCircuitAndRegisters = tuple[
    stim.Circuit, dict[Channel, tuple[str, stim.GateTarget]], dict[Channel, int]
]


StimCallable = Callable[
    [
        list[
            list[stim.GateTarget]
        ],  # List of targets for one stim instruction (e.g. [[0,1], [2,3]] for two CX)
    ],
    list[str],  # Stim instructions strings
]
StimCallable.__doc__ = """
    Type alias that describes a callable that maps an Eka op to stim instruction strings 
    with targets.
    Takes a list of list of targets, allowing multiple instructions of the 
    same type to be grouped.
    For example, two CX gates on different pair of targets:
    [[q1, c1], [q2, c2]] → "CX q1 c1 q2 c2".
    Returns a list of stim instruction strings with targets, since one Eka instruction
    can correspond to multiple stim instructions.
"""

# Type of the run result from stim.sample (NDArray of shape (shots, num_qubits)),
# where each entry is a boolean, holding the measurement outcome of the latest measure
# of each qubits.
StimOutputRunResult = tuple[
    NDArray[np.bool], dict[Channel, tuple[str, stim.GateTarget]]
]


class EkaToStimConverter(Converter[StimCircuitAndRegisters, StimOutputRunResult]):
    """
    Convert an InterpretationStep to a stim.Circuit.

    Here's a simple example of how to use this method to execute Eka experiment in
    Stim:

    .. code-block:: python

        from loom.executor import EkaToStimConverter

        stim_converter = EkaToStimConverter()
        stim_program, quant_register, class_register = stim_converter.convert(
            eka
        )

        import stim
        st = stim_program
        sampler = st.compile_sampler()
        result = sampler.sample(shots=5)

        print(stim_converter.parse_target_run_outcome((result, class_register)))

    """

    # Support only Clifford operations
    SUPPORTED_OPERATIONS: frozenset[OpSignature] = CLIFFORD_GATES_SIGNATURE

    SUPPORTED_ERROR_TYPES: frozenset[ErrorType] = frozenset(
        {
            ErrorType.PAULI_X,
            ErrorType.PAULI_Y,
            ErrorType.PAULI_Z,
            ErrorType.PAULI_CHANNEL,
            ErrorType.BIT_FLIP,
            ErrorType.PHASE_FLIP,
            ErrorType.DEPOLARIZING1,
            ErrorType.DEPOLARIZING2,
        }
    )

    ALLOW_ERROR_MODELS: bool = Field(default=True, frozen=True, init=False)

    # Special classically controlled operations supported by Stim, only allow
    # conditional pauli operations.
    STIM_CLASSICALLY_CONTROLLED_OPS: frozenset[str] = frozenset(
        {
            OpSignature(
                name=n,
                op_type=OpType.CUSTOM,
                classical_input=1,
                quantum_input=1,
            )
            for n in {
                f"classical_controlled_{applied_pauli}"
                for applied_pauli in ["x", "y", "z"]
            }
        }
    )

    @cached_property
    # pylint: disable=unused-argument
    def operations_map(
        self,
    ) -> dict[str, StimCallable]:
        """Map of operation signatures to their corresponding StimCallable, which is a
        set of stim instructions (as strings) associated with values for the targets and
        parameters."""

        def _make_stim_instruction_string(
            name: str,
            targets: list[list[stim.GateTarget]],
            gate_args: list[float] | None = None,
        ) -> str:
            """
            Construct a Stim instruction string for a gate.

            Arguments
            ---------
            name : str
                Gate name (e.g., "X", "H", "CX").
            targets : list[list[stim.GateTarget]]
                Nested list of stim.GateTarget to be flattened.
            gate_args : list[float] | None
                Optional list of float parameters.

            Returns
            -------
            str
                A single Stim instruction string.
            """
            # Flatten targets and convert to Stim string representation
            flattened_targets = [t for sublist in targets for t in sublist]
            targets_str = " ".join(str(t.value) for t in flattened_targets)

            # Handle optional gate arguments
            if gate_args:
                args_str = ", ".join(map(str, gate_args))
                return f"{name}({args_str}) {targets_str}".strip()
            return f"{name} {targets_str}".strip()

        def _handle_op(
            targets: list[list[stim.GateTarget]],
            op,
            gate_args: list[float] | None = None,
        ) -> list[str]:
            """Sort the targets and call the corresponding operation handler.
            Also handle the case where op is a tuple of operations by creating a list of
            instructions.
            """
            if isinstance(op, tuple):
                return [
                    _make_stim_instruction_string(o, targets, gate_args=gate_args)
                    for o in op
                ]
            return [_make_stim_instruction_string(op, targets, gate_args=gate_args)]

        def _control_flow_op(
            targets: list[list[stim.GateTarget]],
            op: str,
            gate_args: list[float] | None = None,
        ) -> list[str]:
            """Handle control flow operations in stim. Only allow conditional paulis.
            targets are used as follow: [[control, target]]

            """
            pauli = op.replace("classical_controlled_", "")
            return [f"C{pauli.upper()} rec[{targets[0][0]}] {str(targets[0][1].value)}"]

        eka_to_stim_ops = {
            "i": "I",
            "hadamard": "H",
            "h": "H",
            "phase": "S",
            "phaseinv": "S_DAG",
            "x": "X",
            "y": "Y",
            "z": "Z",
            "cx": "CX",
            "cnot": "CX",
            "cz": "CZ",
            "cy": "CY",
            "swap": "SWAP",
            "reset": "R",
            "reset_0": "R",
            "reset_1": ("R", "X"),
            "reset_+": "RX",
            "reset_-": ("RX", "Z"),
            "reset_+i": "RY",
            "reset_-i": ("RY", "Z"),
            "measurement": "M",
            "measure_z": "M",
            "measure_x": "MX",
            "measure_y": "MY",
            ErrorType.PAULI_X.label: "X_ERROR",
            ErrorType.PAULI_Y.label: "Y_ERROR",
            ErrorType.PAULI_Z.label: "Z_ERROR",
            ErrorType.BIT_FLIP.label: "X_ERROR",
            ErrorType.PHASE_FLIP.label: "Z_ERROR",
            ErrorType.PAULI_CHANNEL.label: "PAULI_CHANNEL_1",
            ErrorType.DEPOLARIZING1.label: "DEPOLARIZE1",
            ErrorType.DEPOLARIZING2.label: "DEPOLARIZE2",
        }

        return (
            {
                op.name: partial(_handle_op, op=eka_to_stim_ops[op.name])
                for op in CLIFFORD_GATES_SIGNATURE
            }
            | {
                error_type.label: partial(  # type: ignore
                    _handle_op, op=eka_to_stim_ops[error_type.label]
                )
                for error_type in self.SUPPORTED_ERROR_TYPES
            }
            | {
                op.name: partial(_control_flow_op, op=op.name)
                for op in self.STIM_CLASSICALLY_CONTROLLED_OPS
            }
        )

    @staticmethod
    def parse_if_operation(if_circuit: IfElseCircuit) -> Circuit:
        """
        Parse control flow operation, allowing only specific cases supported by Stim.
        Stim only supports conditional pauli operations, with single bit conditions.

        Parameters
        ----------
        if_circuit : IfElseCircuit
            The IfElseCircuit to parse.

        Returns
        -------
        Circuit
            A Circuit (gate) representing the parsed if operation in Stim, which is a
            CX (resp. CY or CZ) operation with a classical channel as control.
        """
        if not if_circuit.is_condition_single_bit:
            raise ValueError(
                "Unsupported operation for Stim conversion: "
                "Stim only supports single bit conditions for if-else circuits."
            )

        if not if_circuit.is_single_gate_conditioned:
            raise ValueError(
                "Unsupported operation for Stim conversion: "
                "Stim only supports single gate conditioned if-else circuits."
            )

        applied_pauli = if_circuit.if_circuit.circuit[0][0].name
        if applied_pauli not in {
            "x",
            "y",
            "z",
        }:
            raise ValueError(
                "Unsupported operation for Stim conversion: "
                "Stim only supports conditional pauli operations in if-else circuits."
            )

        return Circuit(
            name=f"classical_controlled_{applied_pauli}",
            channels=[
                if_circuit.condition_circuit.channels[0],
                if_circuit.if_circuit.channels[0],
            ],
        )

    # pylint: disable=too-many-locals, too-many-nested-blocks, too-many-statements,
    # pylint: disable=too-many-branches
    def emit_circuit_program(
        self,
        input_circuit: Circuit,
        error_models: list[CircuitErrorModel] | None = None,
        with_ticks: bool = False,
    ) -> tuple[
        str, dict[Channel, tuple[str, stim.GateTarget]], dict[Channel, int], int
    ]:
        """Convert a Circuit to a text in stim format and
        a mapping of channel labels to stim target or measurement record indices.

        Parameters
        ----------
        input_circuit : Circuit
            The EKA circuit to convert.
        with_ticks : bool, optional
            Whether to include TICK instructions in the stim circuit, by default False.

        Returns
        -------
        str, dict[Channel, tuple[str, stim.GateTarget]], dict[Channel, int]
            A tuple containing the stim circuit as a string, a dictionary mapping
            quantum channel ids to stim qubit targets and coordinates and a dictionary
            mapping classical channel ids to stim measurement record indices,
        """
        if not isinstance(input_circuit, Circuit):
            raise TypeError("Input must be a Circuit instance.")

        # Initialize an empty stim circuit
        final_instructions = []

        # If the circuit is empty, return the empty stim circuit and empty mappings
        if (
            not input_circuit.channels
            or not input_circuit.circuit
            or len(input_circuit.circuit) == 0
        ):
            return stim.Circuit("\n".join(final_instructions)), {}, {}, 0

        init_instructions, q_chan_to_stim_qubits_map, c_channel_to_rec_idx = (
            self.emit_init_instructions(input_circuit)
        )

        for init_instr in init_instructions.split("\n"):
            if init_instr:  # avoid adding empty lines
                final_instructions.append(init_instr)

        @dataclass(frozen=True)
        class NoiseArgsKey:
            """This class is used to provide a hashable key for information that defines
            a noise instruction."""

            e_type: ErrorType
            a_mode: ApplicationMode
            prob: tuple[float]

        # For ease of access, define the set of classically controlled op names
        classically_controlled_op_names = {
            o.name for o in self.STIM_CLASSICALLY_CONTROLLED_OPS
        }

        # Group operations (=gates) in a layer (tick) by their type
        # stim_instructions_by_layer is a list of dicts, where each dict corresponds to
        # a tick, and the dict map operations name to their targets
        stim_instructions_by_layer: list[
            dict[
                tuple[str, tuple[NoiseArgsKey, ...]],
                list[tuple[list[stim.GateTarget], list[Channel]]],
            ]
        ] = []

        for layer in Circuit.unroll(input_circuit):

            # dict mapping (op_name, *NoiseArgsKey) to list of quantum targets and
            # classical channels
            # This is useful to group operations of the same type in one stim
            # instruction, Including associated noise instructions (if any).
            op_dict = defaultdict(list)
            for op in layer:
                if op.name == "if-else_circuit":
                    op = self.parse_if_operation(op)
                if op.name not in self.operations_map:
                    raise KeyError(f"Operation {op.name} not found in operations map.")
                classical_channels = [c for c in op.channels if c.is_classical()]
                targets = [
                    q_chan_to_stim_qubits_map[c][1]
                    for c in op.channels
                    if c.is_quantum()
                ]

                # Validation of operation arguments
                if op.name in classically_controlled_op_names:
                    if len(classical_channels) != 1 or len(targets) != 1:
                        raise ValueError(
                            f"Classically controlled operation {op.name} must have one "
                            "classical control channel and one quantum target channel."
                        )
                else:
                    self._validate_ops_args(
                        op.name, len(targets), len(classical_channels)
                    )

                noises = []
                for cem in error_models or []:
                    noise_args = cem.get_gate_error_probability(op)
                    if noise_args:
                        noises.append(
                            NoiseArgsKey(
                                e_type=cem.error_type,
                                a_mode=cem.application_mode,
                                prob=tuple(noise_args),
                            )
                        )

                # Group by operation name and associated noise instructions
                op_dict[(op.name, tuple(noises))].append((targets, classical_channels))
            # Append the dict of operations for this layer
            stim_instructions_by_layer.append(op_dict)

        # Keep track of the measurement record index to map them to classical channels
        measurement_record_counter = 0
        tick_count = 0
        # Build the stim circuit from the grouped operations
        for s_layer in stim_instructions_by_layer:
            for ops, targets_and_channels in s_layer.items():
                targets, classical_channels = zip(*targets_and_channels)
                # here the op is a tuple of (op_name, *NoiseArgsKey)
                # unpack it to get the op_name and noise_args
                op, noise_instructions = ops

                # Handle classically controlled operations and measurement operations
                if op in classically_controlled_op_names:
                    # Intertwine the classical target with the quantum targets for
                    # classically controlled operations
                    classical_targets = [
                        [
                            c_channel_to_rec_idx[chan] - measurement_record_counter
                            for chan in chan_list
                        ]
                        for chan_list in classical_channels
                    ]
                    # Set quantum target to be:
                    # [
                    #   [classical_targets[0], quantum_targets[0]],
                    #   [classical_targets[1], quantum_targets[1]],
                    #  ...
                    # ]
                    targets = [
                        [t_pair[0][0], t_pair[1][0]]
                        for t_pair in zip(classical_targets, targets)
                    ]
                elif self.op_by_eka_name[op].op_type == OpType.MEASUREMENT:
                    # Map classical channels to measurement record indices
                    # for measurement operations
                    for c in classical_channels:
                        c_channel_to_rec_idx[c[0]] = measurement_record_counter
                        measurement_record_counter += 1

                # Get the stim instructions for this operation
                stim_instructions = self.operations_map[op](targets)
                # Add noise instructions associated with this operation
                for noise in noise_instructions:
                    if noise.a_mode == ApplicationMode.BEFORE_GATE:
                        stim_instructions = (
                            self.operations_map[noise.e_type.label](
                                targets=targets, gate_args=noise.prob
                            )
                            + stim_instructions
                        )
                    if noise.a_mode == ApplicationMode.AFTER_GATE:
                        stim_instructions = stim_instructions + self.operations_map[
                            noise.e_type.label
                        ](targets=targets, gate_args=noise.prob)

                final_instructions.extend(stim_instructions)

            # Add tick-wise noise if error model is provided
            noise_instruction = []
            for cem in error_models or []:
                if cem.application_mode == ApplicationMode.END_OF_TICK:
                    noise_args = cem.get_tick_error_probability(tick_count)
                    noise_instruction += self.operations_map[cem.error_type.label](
                        targets=[
                            [
                                value[1]
                                for key, value in q_chan_to_stim_qubits_map.items()
                            ]
                        ],
                        gate_args=noise_args,
                    )
                if cem.application_mode == ApplicationMode.IDLE_END_OF_TICK:
                    # group all the target with the same error probabilities together
                    idle_args = {}
                    for c, stim_reg in q_chan_to_stim_qubits_map.items():
                        args = cem.get_idle_tick_error_probability(tick_count, c.id)
                        if args:
                            if tuple(args) in idle_args:
                                idle_args[tuple(args)].append(stim_reg[1])
                            else:
                                idle_args[tuple(args)] = [stim_reg[1]]
                    for args, targets in idle_args.items():
                        noise_instruction += self.operations_map[cem.error_type.label](
                            targets=[targets], gate_args=args
                        )

            final_instructions.extend(noise_instruction)

            tick_count += 1
            # Append tick instruction for every eka layer
            if with_ticks is True:
                final_instructions.append("TICK")

        return (
            final_instructions,
            q_chan_to_stim_qubits_map,
            c_channel_to_rec_idx,
            measurement_record_counter,
        )

    def convert_circuit(
        self,
        input_circuit: Circuit,
        error_models: list[CircuitErrorModel] | None = None,
        with_ticks: bool = False,
    ):
        """Convert a Circuit to a text in stim format and
        a mapping of channel labels to stim target or measurement record indices.

        Parameters
        ----------
        input_circuit : Circuit
            The EKA circuit to convert.
        with_ticks : bool, optional
            Whether to include TICK instructions in the stim circuit, by default False.

        Returns
        -------
        StimCircuitAndRegisters
            A tuple containing the stim circuit, a dictionary mapping quantum channel
            ids to stim qubit targets and coordinates and a dictionary mapping
            classical channel ids to stim measurement record indices,
        """
        if not input_circuit.circuit or len(input_circuit.circuit) == 0:
            return "", {}, {}
        circ_text, quant_dict, rec_dict, _ = self.emit_circuit_program(
            input_circuit, error_models=error_models, with_ticks=with_ticks
        )
        return stim.Circuit("\n".join(circ_text)), quant_dict, rec_dict

    def convert(
        self,
        interpreted_eka: InterpretationStep,
        error_models: list[CircuitErrorModel] | None = None,
        with_ticks: bool = False,
    ) -> StimCircuitAndRegisters:
        """
        Convert a InterpretationStep.
        Call the default convert method to convert the final circuit of the step.
        Then, add the detectors and logical observables to the stim circuit.
        """

        # Convert the final circuit of the interpretation step
        input_circuit = interpreted_eka.final_circuit
        if not input_circuit.circuit or len(input_circuit.circuit) == 0:
            return stim.Circuit(), {}, {}
        stim_circuit_text, quant_dict, rec_dict, measurement_counter = (
            self.emit_circuit_program(
                input_circuit, error_models=error_models, with_ticks=with_ticks
            )
        )
        # Get and add detectors to the stim circuit
        detector_instructions = EkaToStimConverter.emit_detectors_and_observables(
            interpreted_eka, rec_dict, measurement_counter
        )
        if detector_instructions:
            stim_circuit_text.extend(detector_instructions)

        return stim.Circuit("\n".join(stim_circuit_text)), quant_dict, rec_dict

    @staticmethod
    def parse_target_run_outcome(
        output: StimOutputRunResult,
    ) -> dict[str, int | list[int]]:
        """
        Parse the run output of a stim circuit into a dictionary mapping the
        Eka classical channels labels to their measurement outcomes.

        Parameters
        ----------
        output : StimOutputRunResult
            The output from stim simulation, which is a tuple of
            (NDArray of shape (shots, num_qubits), dict mapping classical channel to
            stim measurement record index).

        Returns
        -------
        dict[Channel, int | list[int]]
            A dictionary mapping the Eka classical channel labels to their measurement
            outcomes. If multiple shots were run, the outcomes are lists of integers
            (0 or 1), otherwise a single integer (0 or 1).
        """
        output, c_reg = output
        res: dict[Channel, int | list[int]] = {}
        n_shots = output.shape[0]

        for key, idx in c_reg.items():
            values = output[:, idx].astype(int).tolist()
            # Flatten to int if only one shot
            res[key.label] = values[0] if n_shots == 1 else values

        return res

    @staticmethod
    # pylint: disable=too-many-return-statements
    def _eka_to_stim_coordinates(
        eka_coord: tuple[int, ...] | str,
    ) -> int | tuple[float, ...]:
        """
        Convert EKA coordinates to Stim coordinates. If possible (i.e. if the label of
        the quantum channel
        is in the form of (x, y, z) or (x, y)), otherwise return 0
        """
        if isinstance(eka_coord, tuple):
            coords = eka_coord
        else:
            try:
                coords = ast.literal_eval(eka_coord)
            except (ValueError, SyntaxError):
                coords = eka_coord  # treat it as a raw string

        if not isinstance(coords, tuple):
            return 0

        match len(coords):
            case 2:
                # For linear lattice codes
                if coords[1] == 1:
                    return (coords[0], 0)
                if coords[1] == 0:
                    return (coords[0] - 0.5, 0.5)
                # Removed error and kept this temporarily to host HGP codes until
                # Lattice refactor
                return coords
            case 3:
                # For square lattice codes
                if coords[2] == 1:
                    return (coords[0], coords[1])
                if coords[2] == 0:
                    return (coords[0] + 0.5, coords[1] + 0.5)
                # Patched up case for proper handling of Color Codes until Lattice
                # refactor
                if coords[2] == 2:
                    return (coords[0] + 0.05, coords[1] + 0.05)
                raise ValueError(
                    f"Invalid coordinate {coords}. "
                    "Coordinates should be in the form (x, y, 0), (x, y, 1)"
                    " or (x, y, 2)."
                )
            case _:
                # For other lattice codes raise an error
                raise ValueError(f"Invalid channel label {eka_coord}.")

    @staticmethod
    def emit_detectors_and_observables(
        input_istep: InterpretationStep,
        rec_dict: dict[str, int],
        measurement_offset: int = 0,
    ) -> list[str]:
        """Get the detector instructions for the given interpretation step, and the
        given measurement offset which corresponds to the number of measurements in the
        circuit before adding detectors.

        Parameters
        ----------
        input_istep : InterpretationStep
            The interpretation step to get detector instructions for.
        rec_dict : dict[str, int]
            A mapping from channel labels to their corresponding record indices.
        measurement_offset : int, optional
            An offset to apply to the measurement indices, by default 0

        Returns
        -------
        list[str]
            A list of stim circuit instructions for the detector measurements.

        Raises
        ------
        ValueError
            If the detector labels are not in the expected format.
        """
        detector_instructions = []
        # Add detectors to the stim circuit
        for det in input_istep.detectors:
            # Find the channels corresponding to the measurements in the detector
            # According to some obscure convention.
            channels_in_det = [
                channel
                for meas in det.measurements
                for channel in input_istep.final_circuit.channels
                if isinstance(meas, tuple) and channel.label == f"{meas[0]}_{meas[1]}"
                # ignore constant cbits with the isinstance check
            ]

            # substract the number of measurements in the circuit to get the
            # correct indices in the stim circuit (which track records with negative
            # indices, -1 is the latest measurement, -2 the one before, etc.)
            targets = [
                f"rec[{rec_dict[chan] - measurement_offset}]"
                for chan in channels_in_det
            ]

            # Eze's shinegan, I have no idea why this works, this follows obscure
            # conventions again.
            det_args = []
            try:
                if (
                    "space_coordinates" in det.labels
                    and "time_coordinate" in det.labels
                ):
                    det_args = list(
                        EkaToStimConverter._eka_to_stim_coordinates(
                            det.labels.get("space_coordinates")
                        )
                        + det.labels.get("time_coordinate")
                    )
                if "color" in det.labels:
                    det_args.append(det.labels.get("color"))

            except Exception as e:
                traceback.print_exc()
                raise e
            if det_args:
                det_args = [str(arg) for arg in det_args]
                str_instruction = f"DETECTOR({', '.join(det_args)}) {' '.join(targets)}"
            else:
                str_instruction = f"DETECTOR {' '.join(targets)}"

            detector_instructions.append(str_instruction)

        # Add observables to the stim circuit
        for i, logical_observable in enumerate(input_istep.logical_observables):
            # ignore constant cbits with the isinstance check
            target_channel = [
                channel
                for meas in logical_observable.measurements
                for channel in input_istep.final_circuit.channels
                if isinstance(meas, tuple) and channel.label == f"{meas[0]}_{meas[1]}"
            ]

            # Get the targets in stim format from the measurement record indices
            targets = [
                f"rec[{rec_dict[chan] - measurement_offset}]" for chan in target_channel
            ]
            detector_instructions.append(
                f"OBSERVABLE_INCLUDE({str(i)}) {' '.join(targets)}"
            )

        return detector_instructions

    def emit_init_instructions(
        self, circuit: Circuit
    ) -> tuple[str, dict[Channel, Any], dict[Channel, Any]]:
        """Provide the python code (as a string) to initializes the
        quantum and classical registers, and return the mapping from eka channel to
        register.

        In the beginning of the circuit, the mapping of classical channels to
        measurement record indices is None, since no classical channels have been
        measured yet."""

        if not isinstance(circuit, Circuit):
            raise TypeError("Input must be a Circuit instance.")

        instructions = []

        # Sort quantum channels by label to ensure consistent qubit ordering across runs
        q_channels = sorted(
            [c for c in circuit.channels if c.is_quantum()],
            key=lambda x: x.label,
        )

        # Map of EKA qubits to Stim qubits idx and coordinates
        q_chan_to_stim_qubits_map = {}

        # Create mapping of Eka quantum channels to stim qubit indices
        # and add QUBIT_COORDS instructions to the stim circuit if coordinates
        # are provided.
        for i, chan in enumerate(q_channels):
            stim_coords = EkaToStimConverter._eka_to_stim_coordinates(chan.label)
            if stim_coords != 0:
                q_chan_to_stim_qubits_map[chan] = (stim_coords, stim.GateTarget(i))
                instructions.append(f"QUBIT_COORDS{stim_coords} {i}")
            else:
                # If the label is not a coordinate, just map to the qubit index
                q_chan_to_stim_qubits_map[chan] = (None, stim.GateTarget(i))

        # Create mapping of Eka classical channels to stim measurement record indices
        # (None if no classical channels have not been measured yet)
        c_channels = [c for c in circuit.channels if c.is_classical()]
        c_channel_to_rec_idx = {c: None for c in c_channels}

        return (
            "\n".join(instructions),
            q_chan_to_stim_qubits_map,
            c_channel_to_rec_idx,
        )

    def emit_leaf_circuit_instruction(
        self,
        circuit: Circuit,
        quantum_channel_map: dict[Channel, tuple[Any, stim.GateTarget]],
        classical_channel_map: dict[Channel, Any],
        measurement_record_counter: int = 0,
    ) -> str:
        """Provide the python code (as a string) to emit an Eka instruction in the
        target language."""

        if not isinstance(circuit, Circuit):
            raise TypeError("Input must be a Circuit instance.")

        if circuit.circuit and len(circuit.circuit) != 0:
            raise ValueError("The circuit must be an leaf circuit with channels.")

        instructions = []
        targets = [
            [quantum_channel_map[c][1]] for c in circuit.channels if c.is_quantum()
        ]
        classical_target = [c for c in circuit.channels if c.is_classical()]
        # Build the stim circuit from the grouped operations

        self._validate_ops_args(circuit.name, len(targets), len(classical_target))

        if circuit.name in self.operations_map:
            stim_instructions = self.operations_map[circuit.name](targets)
            for instr in stim_instructions:
                instructions.append(str(instr))

        if self.op_by_eka_name[circuit.name].op_type == OpType.MEASUREMENT:
            classical_channel_map[classical_target[0]] = measurement_record_counter
            measurement_record_counter += 1

        return "\n".join(instructions)

    @classmethod
    def stim_polygons(cls, interpreted_eka: InterpretationStep) -> str:
        """Define stim polygons using data qubits coordinates involved
        with each stabilizer on patches passed as argument to the function

        DEMO SYNTAX: #!pragma POLYGON(1,0,0,0.25) 5 11 16 23
            POLYGON(<X>, <Y>, <Z>, <color intensity>) <data qubits involved>

        Since polygon definitions are added as comments in the stim circuit body,
        and there is no way to add comments programmatically in stim.Circuit
        This function is only available to print polygon instructions from the
        patch stabilizers. The user *MUST* add these comments manually to the
        stim.Circuit string body to display the polygons in crumble

        Parameters
        ----------
        interpreted_eka: InterpretationStep
            The `InterpretationStep` object containing
            information on the code stabilizers

        Returns
        -------
        stim_polygons: str
            Stim polygon instructions as a string
        """
        # list of stabilizers to define the polygons
        pauli_polyarg_map = {
            "X": "(1,0,0,0.5)",
            "Y": "(0,1,0,0.5)",
            "Z": "(0,0,1,0.5)",
        }
        # Get all blocks at the initial timestamp (timestamp 0)
        initial_block_uuids = interpreted_eka.block_history.blocks_at(0)
        all_stabilizers = [
            stab
            for block_uuid in initial_block_uuids
            for stab in interpreted_eka.block_registry[block_uuid].stabilizers
        ]

        eka_circuit = interpreted_eka.final_circuit
        eka_channels = [
            ast.literal_eval(chan.label)
            for chan in eka_circuit.channels
            if chan.is_quantum()
        ]

        eka_coord_to_stim_qubit_map = {
            coords: stim.CircuitInstruction(
                name="QUBIT_COORDS",
                targets=[i],
                gate_args=cls._eka_to_stim_coordinates(coords),
            )
            for i, coords in enumerate(
                sorted(eka_channels, key=cls._eka_to_stim_coordinates)
            )
        }

        polygon_instructions = []
        for stab in all_stabilizers:
            if len(set(list(stab.pauli))) == 1:
                polyarg = pauli_polyarg_map[stab.pauli[0]]
                data_qubits = [
                    eka_coord_to_stim_qubit_map[data_qubit]
                    for data_qubit in stab.data_qubits
                ]
                # arrange qubits in cyclically to be visualized properly by crumble
                polygon_ordered_data_qubits = (
                    data_qubits[int(len(data_qubits) / 2) :]
                    + data_qubits[: int(len(data_qubits) / 2)][::-1]
                )
            else:
                raise ValueError(
                    f"Unsupported {stab.pauli}."
                    "Currently only CSS type codes are supported."
                )
            polygon_instructions.append(
                f"#!pragma POLYGON{polyarg} "
                + " ".join(
                    str(qubit.targets_copy()[0].value)
                    for qubit in polygon_ordered_data_qubits
                )
            )
            polygon_instructions_string = "\n".join(
                instruction for instruction in polygon_instructions
            )
        return polygon_instructions_string

    def print_stim_circuit_for_crumble(self, final_step: InterpretationStep) -> str:
        """Print the stim circuit along with polygon instructions to be
        used for crumble

        Parameters
        ----------
        stim_circ : stim.Circuit
            input stim circuit
        """
        polygon_instructions = self.stim_polygons(final_step)
        stim_circuit, _, _ = self.convert(final_step)

        total_output = polygon_instructions + "\n" + str(stim_circuit)

        return total_output
